from math import ceil
import string
import re
from typing import Tuple, Optional, List


def _get_squares(variant: Optional[str]) -> Tuple[int, int, int, bool]:
    """Returns the total squares, squares per row, squares per column and if every other square is playable."""
    # The default values are for International draughts.
    total_squares = 50
    squares_per_row = 5
    squares_per_column = 5
    every_other_square = True
    if variant in ['english', 'italian', 'russian', 'brazilian']:
        total_squares = 32
        squares_per_row = 4
        squares_per_column = 4
    elif variant in ['turkish']:
        total_squares = 64
        squares_per_row = 8
        squares_per_column = 8
        every_other_square = False
    return total_squares, squares_per_row, squares_per_column, every_other_square


def _rotate_move(internal_move: str, notation: Optional[int] = None, variant: Optional[str] = None) -> str:
    """Rotate the move."""
    separators = ['-', 'x', ':']
    splitted_move = None
    correct_seperator = ''
    for separator in separators:
        splitted_move = internal_move.split(separator)
        if splitted_move[0] != internal_move:
            correct_seperator = separator
            break
    splitted_move = list(map(int, splitted_move))
    variant_to_notation = {'standard': 2, 'english': 1, 'italian': 2, 'russian': 0, 'brazilian': 0, 'turkish': 0, 'frisian': 2, 'frysk!': 2, 'antidraughts': 2, 'breakthrough': 2}
    if notation is None:
        notation = variant_to_notation.get(variant, 2)

    def reverse_column(splitted_int_move: List[int]) -> List[int]:
        per_row = _get_squares(variant)[1]
        for index, square in enumerate(splitted_int_move):
            square_in_row = square % per_row
            if square_in_row == 0:
                square_in_row += per_row
            splitted_int_move[index] = ((square - 1) // per_row) * per_row + (per_row - (square_in_row - 1))
        return splitted_int_move

    def reverse_row_and_column(splitted_int_move: List[int]) -> List[int]:
        squares = _get_squares(variant)[0]
        splitted_int_move = list(map(lambda square: squares + 1 - square, splitted_int_move))
        return splitted_int_move

    def reverse_row(splitted_int_move: List[int]) -> List[int]:
        return reverse_column(reverse_row_and_column(splitted_int_move))

    rotated_move = None
    if notation == 0:
        rotated_move = reverse_row(splitted_move)
    elif notation == 1:
        rotated_move = reverse_row_and_column(splitted_move)
    elif notation == 2:
        rotated_move = splitted_move
    elif notation == 3:
        rotated_move = reverse_column(splitted_move)

    rotated_move = list(map(str, rotated_move))
    return correct_seperator.join(rotated_move)


def _algebraic_to_number(algebraic_move: str, squares_per_letter: Optional[int] = None, variant: Optional[str] = None, every_other_square: Optional[bool] = None) -> str:
    """Convert an algebraic move to a numeric move."""
    if every_other_square is None:
        if variant == 'turkish':
            every_other_square = False
        else:
            every_other_square = True
    algebraic_notation = algebraic_move[0] in string.ascii_letters
    if not algebraic_notation:
        return algebraic_move
    algebraic_move = algebraic_move.lower()
    if squares_per_letter is None:
        squares_per_letter = _get_squares(variant)[2]

    separators = ['-', 'x', ':']
    special_seperators = [r'([a-zA-z]+\d+)']
    splitted_move = None
    correct_seperator = ''
    for separator in separators:
        splitted_move = list(filter(bool, re.split(separator, algebraic_move)))
        if splitted_move[0] != algebraic_move:
            correct_seperator = separator
            break

    if not correct_seperator:
        for separator in special_seperators:
            splitted_move = list(filter(bool, re.split(separator, algebraic_move)))
            if splitted_move[0] != algebraic_move:
                correct_seperator = '-'
                break

    numeric_move = []
    for move in splitted_move:
        numeric_move.append(_algebraic_to_numeric_square(move, squares_per_letter, every_other_square=every_other_square))
    numeric_move = list(map(str, numeric_move))
    return correct_seperator.join(numeric_move)


def _algebraic_to_numeric_square(square: str, squares_per_letter: int, every_other_square: bool = True) -> int:
    """Convert an algebraic square to a numeric square."""
    algebraic_notation = square[0] in string.ascii_letters
    if not algebraic_notation:
        return int(square)
    if not every_other_square:
        return (int(square[1]) - 1) * squares_per_letter + string.ascii_lowercase.index(square[0]) + 1
    return (int(square[1]) - 1) * squares_per_letter + ceil(string.ascii_lowercase.index(square[0]) // 2) + 1


def _number_to_algebraic(number_move: str, width: Optional[int] = None, variant: Optional[str] = None, every_other_square: Optional[bool] = None) -> str:
    """Convert a numeric move to an algebraic move."""
    if every_other_square is None:
        if variant == 'turkish':
            every_other_square = False
        else:
            every_other_square = True
    algebraic_notation = number_move[0] in string.ascii_letters
    if algebraic_notation:
        return number_move
    if width is None:
        width = _get_squares(variant)[1]

    separators = ['-', 'x', ':']
    splitted_move = None
    correct_seperator = ''
    for separator in separators:
        splitted_move = list(filter(bool, re.split(separator, number_move)))
        if splitted_move[0] != number_move:
            correct_seperator = separator
            break

    algebraic_move = []
    for move in splitted_move:
        algebraic_move.append(_numeric_to_algebraic_square(move, width, every_other_square=every_other_square))
    return correct_seperator.join(algebraic_move)


def _numeric_to_algebraic_square(square: str, width: int, every_other_square: Optional[bool] = True) -> str:
    """Convert a numeric square to an algebraic square."""
    algebraic_notation = square[0] in string.ascii_letters
    if algebraic_notation:
        return square
    square = int(square)
    row = ceil(square / width) - 1
    column = (square - 1) % width
    if every_other_square:
        column *= 2
        column += 1 if row % 2 == 1 else 0
    return string.ascii_lowercase[column] + str(row + 1)


def _change_fen_from_variant(li_fen: str, notation: Optional[int] = None, squares_per_letter: int = 5, every_other_square: bool = True, variant: Optional[str] = None) -> str:
    """Convert an internal fen to the correct fen for the variant."""
    if variant:
        _, _, squares_per_letter, every_other_square = _get_squares(variant)

    li_fen = li_fen.split(':')
    starts = li_fen[0]
    white_pieces = li_fen[1][1:].split(',')
    black_pieces = li_fen[2][1:].split(',')

    white_pieces_remove_hyphen = []
    for white_piece in white_pieces:
        if '-' in white_piece:
            start_end = white_piece.split('-')
            add_for_king = ''
            if start_end[0][0] == 'K':
                add_for_king = 'K'
                start_end[0] = start_end[0][1:]
            start, end = _algebraic_to_numeric_square(start_end[0], squares_per_letter, every_other_square), _algebraic_to_numeric_square(start_end[1], squares_per_letter, every_other_square)
            for number in range(start, end + 1):
                white_pieces_remove_hyphen.append(add_for_king + _rotate_move(str(number), notation=notation, variant=variant))
        else:
            add_for_king = ''
            if white_piece[0] == 'K':
                add_for_king = 'K'
                white_piece = white_piece[1:]
            white_pieces_remove_hyphen.append(add_for_king + _rotate_move(str(_algebraic_to_numeric_square(white_piece, squares_per_letter, every_other_square)), notation=notation, variant=variant))

    black_pieces_remove_hyphen = []
    for black_piece in black_pieces:
        if '-' in black_piece:
            start_end = black_piece.split('-')
            add_for_king = ''
            if start_end[0][0] == 'K':
                add_for_king = 'K'
                start_end[0] = start_end[0][1:]
            start, end = _algebraic_to_numeric_square(start_end[0], squares_per_letter, every_other_square), _algebraic_to_numeric_square(start_end[1], squares_per_letter, every_other_square)
            for number in range(start, end + 1):
                black_pieces_remove_hyphen.append(add_for_king + _rotate_move(str(number), notation=notation, variant=variant))
        else:
            add_for_king = ''
            if black_piece[0] == 'K':
                add_for_king = 'K'
                black_piece = black_piece[1:]
            black_pieces_remove_hyphen.append(add_for_king + _rotate_move(str(_algebraic_to_numeric_square(black_piece, squares_per_letter, every_other_square)), notation=notation, variant=variant))

    # Because in english black starts.
    white_starts = variant not in ['english']
    if not white_starts:
        white_pieces_remove_hyphen, black_pieces_remove_hyphen = black_pieces_remove_hyphen, white_pieces_remove_hyphen
        starts = 'W' if starts == 'B' else 'B'

    white_pieces_remove_hyphen = list(map(lambda move: _algebraic_to_numeric_square(move, squares_per_letter) if move[0].lower() != 'k' else move, white_pieces_remove_hyphen))
    black_pieces_remove_hyphen = list(map(lambda move: _algebraic_to_numeric_square(move, squares_per_letter) if move[0].lower() != 'k' else move, black_pieces_remove_hyphen))
    white_pieces_remove_hyphen = list(map(str, white_pieces_remove_hyphen))
    black_pieces_remove_hyphen = list(map(str, black_pieces_remove_hyphen))
    white_pieces_remove_hyphen.sort()
    black_pieces_remove_hyphen.sort()
    return f'{starts}:W{",".join(white_pieces_remove_hyphen)}:B{",".join(black_pieces_remove_hyphen)}'


def fen_from_variant(fen, variant=None, notation=None, squares_per_letter=None, every_other_square=None):
    """Convert variant fen to internal fen."""
    variant = variant.lower() if variant else variant
    fen = _change_fen_from_variant(fen, variant=variant, notation=notation, squares_per_letter=squares_per_letter, every_other_square=every_other_square)
    return fen


def fen_to_variant(fen, variant=None, notation=None, width=None, squares_per_letter=None, every_other_square=None, to_algebraic=None):
    """Convert internal fen to variant fen."""
    variant = variant.lower() if variant else variant
    fen = _change_fen_from_variant(fen, variant=variant, notation=notation, squares_per_letter=squares_per_letter, every_other_square=every_other_square)
    if to_algebraic or variant in ['russian', 'brazilian', 'turkish']:
        new_white_pieces = []
        new_black_pieces = []
        white_pieces = fen.split(':')[1][1:]
        black_pieces = fen.split(':')[2][1:]
        for piece in white_pieces.split(','):
            add = ''
            if piece.lower().startswith('k'):
                add = 'K'
                piece = piece[1:]
            new_white_pieces.append(add + _number_to_algebraic(piece, variant=variant, width=width, every_other_square=every_other_square))
        for piece in black_pieces.split(','):
            add = ''
            if piece.lower().startswith('k'):
                add = 'K'
                piece = piece[1:]
            new_black_pieces.append(add + _number_to_algebraic(piece, variant=variant, width=width, every_other_square=every_other_square))
        fen = f'{fen[0]}:W{",".join(new_white_pieces)}:B{",".join(new_black_pieces)}'
    return fen


def move_from_variant(move, variant=None, notation=None, squares_per_letter=None, every_other_square=None):
    """Convert variant PDN move to internal PDN move."""
    variant = variant.lower() if variant else variant
    move = _algebraic_to_number(move, variant=variant, squares_per_letter=squares_per_letter, every_other_square=every_other_square)
    move = _rotate_move(move, variant=variant, notation=notation)
    return move


def move_to_variant(move, variant=None, notation=None, width=None, every_other_square=None, to_algebraic=None):
    """Convert internal PDN move to variant PDN move."""
    variant = variant.lower() if variant else variant
    move = _rotate_move(move, variant=variant, notation=notation)
    if to_algebraic or variant in ['russian', 'brazilian', 'turkish']:
        move = _number_to_algebraic(move, variant=variant, width=width, every_other_square=every_other_square)
    return move



from draughts.convert import move_from_variant
from typing import List, Optional, Any


class Move:
    def __init__(self, board: Any = None, board_move: List[List[int]] = None, hub_move: str = None, hub_position_move: str = None, pdn_move: str = None, pdn_position_move: str = None, steps_move: List[int] = None, li_api_move: List[str] = None, li_one_move: str = None, has_captures: Optional[bool] = None, possible_moves: List[List[List[int]]] = None, possible_captures: List[List[Optional[int]]] = None, hub_to_pdn_pseudolegal: bool = False, variant: Optional[str] = None, notation: int = 2, squares_per_letter: int = 4) -> None:
        self.board_move = board_move
        self.hub_move = hub_move
        self.hub_position_move = hub_position_move
        self.pdn_move = pdn_move
        self.pdn_position_move = pdn_position_move
        self.steps_move = steps_move
        self.li_api_move = li_api_move
        self.li_one_move = li_one_move
        self.ambiguous = None
        self.captures = None
        self.has_captures = has_captures
        self.possible_moves = possible_moves
        self.possible_captures = possible_captures
        self.hub_to_pdn_pseudolegal = hub_to_pdn_pseudolegal
        self.variant = variant
        self.original_pdn_move = self.pdn_move
        self.notation = notation
        self.squares_per_letter = squares_per_letter

        if board_move or hub_move or hub_position_move or pdn_move or pdn_position_move or steps_move or li_api_move or li_one_move:
            if board or possible_moves and possible_captures:
                if not possible_moves or not possible_captures:
                    self.possible_moves, self.possible_captures = board.legal_moves()
                self._to_board()
                self.captures = self.possible_captures[self.possible_moves.index(self.board_move)]
                self.captures = [] if self.captures[0] is None else self.captures
                self.has_captures = bool(self.captures)
                self._from_board()
            else:
                self._no_board()

    def _make_len_2(self, move):
        """
        Add a 0 in the front of the square if it is only 1 digit.
        e.g. The move 5 will be turned to 05 but the move 23 will be left the same.
        """
        return f'0{move}' if len(str(move)) == 1 else str(move)

    def _sort_captures(self, captures):
        """
        Sort the captures from the smallest number to the highest.
        e.g. [10, 30, 19] will change to '101930'.
        This function exists because hub engines return the captures in alphabetical order
        (e.g. for the move 231201 scan returns 23x01x07x18 instead of 23x01x18x07)
        """
        if captures and captures[0] is None:
            captures = []
        captures = list(map(self._make_len_2, captures))
        captures.sort()
        captures = ''.join(captures)
        return captures

    def _to_board(self) -> None:
        """Convert the move to all other move types. Requires a Game() object to make the conversions."""

        # Hub related move

        if self.hub_move:
            # Hub move
            if "-" in self.hub_move:
                self.hub_position_move = "".join(list(map(self._make_len_2, self.hub_move.split("-"))))
            else:
                self.hub_position_move = "".join(list(map(self._make_len_2, self.hub_move.split("x"))))

        # PDN related move

        if self.pdn_move:
            # PDN move
            self.pdn_move = move_from_variant(self.pdn_move, variant=self.variant, notation=self.notation, squares_per_letter=self.squares_per_letter)

            if "-" in self.pdn_move:
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.pdn_move.split("-"))))
            else:
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.pdn_move.split("x"))))

        # Hub related move

        if self.hub_position_move:
            # Hub position move
            moves_li_board = {}
            for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                li_move = self._make_len_2(possible_move[0][0]) + self._make_len_2(
                    possible_move[-1][1]) + self._sort_captures(possible_capture)
                moves_li_board[li_move] = possible_move
            board_move = moves_li_board[self.hub_position_move]
            self.board_move = board_move

        # PDN related move

        elif self.pdn_position_move:
            # PDN position move
            moves_li_board = {}
            if len(self.pdn_position_move) == 4:
                self.ambiguous = False
                for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                    li_move = self._make_len_2(possible_move[0][0]) + self._make_len_2(possible_move[-1][1])
                    moves_li_board[li_move] = possible_move
            else:
                self.ambiguous = True
                for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                    steps = [possible_move[0][0]]
                    for move in possible_move:
                        steps.append(move[1])
                    li_move = "".join(list(map(self._make_len_2, steps)))
                    moves_li_board[li_move] = possible_move
            board_move = moves_li_board[self.pdn_position_move]
            self.board_move = board_move

        # Board related moves

        elif self.steps_move:
            # steps_move
            board_move = []
            for index in range(1, len(self.steps_move)):
                board_move.append([self.steps_move[index - 1], self.steps_move[index]])
            self.board_move = board_move

        elif self.li_api_move:
            # li_api_move
            board_move = []
            for move in self.li_api_move:
                board_move.append([int(move[:2]), int(move[2:])])
            self.board_move = board_move

        elif self.li_one_move:
            # li_one_move
            steps = [int(self.li_one_move[i:i + 2]) for i in range(0, len(self.li_one_move), 2)]
            board_move = []
            for index in range(1, len(steps)):
                board_move.append([steps[index - 1], steps[index]])
            self.board_move = board_move

    def _from_board(self) -> None:
        """Convert the move to a board_move. Requires a Game() object to make the conversions."""

        # Board related move

        if not self.steps_move:
            # steps_move
            steps = [self.board_move[0][0]]
            for move in self.board_move:
                steps.append(move[1])
            self.steps_move = steps

        # Hub related moves

        if not self.hub_position_move:
            # Hub position move
            self.hub_position_move = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])
            if self.captures:
                sorted_captures = self._sort_captures(self.captures)
                sorted_captures = [sorted_captures[i:i + 2] for i in range(0, len(sorted_captures), 2)]
                self.hub_position_move += "".join(list(map(self._make_len_2, sorted_captures)))

        if not self.hub_move:
            # Hub move
            positions = [self.hub_position_move[i:i + 2] for i in range(0, len(self.hub_position_move), 2)]
            separator = "x" if self.captures else "-"
            self.hub_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        # PDN related moves

        if not self.pdn_position_move:
            # starts_endings contains the start position and the end position of every possible move.
            # It is used to detect ambiguous PDN moves, so we know when we have to expand the PDN move.
            # e.g. Use 3x14x23 instead of 3x23.

            # PDN position move

            start_end = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])
            same_start_ends = 0
            moves_with_same_start_end_and_captures = []
            for possible_move, possible_captures in zip(self.possible_moves, self.possible_captures):
                possible_start_end = self._make_len_2(possible_move[0][0]) + self._make_len_2(possible_move[-1][1])
                if possible_start_end == start_end:
                    same_start_ends += 1
                    possible_steps_move = [possible_move[0][0]]
                    for semi_move in possible_move:
                        possible_steps_move.append(semi_move[1])
                    moves_with_same_start_end_and_captures.append(possible_steps_move)
            self.ambiguous = same_start_ends >= 2
            if self.ambiguous:
                # From FMJD (https://pdn.fmjd.org/grammar.html#pdn-3-0-grammar):
                # Disambiguated capture sequences have to specify a complete sequence of intermediate squares
                # along the path of the capture. If there is a change in direction, an intermediate square is
                # the square where a turn in direction was made. If there was not a change in direction,
                # the intermediate square is the square immediately behind a captured piece.
                #
                # This means that if we continue going in the same direction,
                # the intermediate square is the square immediately behind the captured piece.
                # pydraughts supports providing a square not immediately behind the captured piece, but when we
                # convert it to a PDN move, we have to change it to the square immediately behind the captured piece.
                correct_start_ends = []
                for index in range(1, len(self.steps_move) - 1):
                    closest_distance = 1000

                    for steps_move in moves_with_same_start_end_and_captures:
                        distance = abs(steps_move[index - 1] - steps_move[index])

                        # New closest square
                        if distance < closest_distance:
                            closest_distance = distance
                            correct_start_ends = [steps_move]
                        # The same intermediate square.
                        # This means that the difference is in a later intermediate square.
                        elif distance == closest_distance:
                            correct_start_ends.append(steps_move)
                        moves_with_same_start_end_and_captures = correct_start_ends

                correct_move = correct_start_ends[0]

                self.pdn_position_move = "".join(list(map(self._make_len_2, correct_move)))
            else:
                self.pdn_position_move = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])

        if not self.pdn_move:
            # PDN move
            positions = [self.pdn_position_move[i:i + 2] for i in range(0, len(self.pdn_position_move), 2)]
            separator = "x" if self.captures else "-"
            self.pdn_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        # Board related moves

        if not self.li_api_move:
            # li_api_move
            li_api_move = []
            for move in self.board_move:
                li_api_move.append(self._make_len_2(move[0]) + self._make_len_2(move[1]))
            self.li_api_move = li_api_move

        if not self.li_one_move:
            # li_one_move
            self.li_one_move = "".join(list(map(self._make_len_2, self.steps_move)))

    def _no_board(self) -> None:
        """Makes as many conversions as possible if the board was not given."""

        # Board related moves

        if self.steps_move:
            # steps_move
            board_move = []
            for index in range(1, len(self.steps_move)):
                board_move.append([self.steps_move[index - 1], self.steps_move[index]])
            self.board_move = board_move

        elif self.li_api_move:
            # li_api_move
            board_move = []
            for move in self.li_api_move:
                board_move.append([int(move[:2]), int(move[2:])])
            self.board_move = board_move

        elif self.li_one_move:
            # li_one_move
            steps = [int(self.li_one_move[i:i + 2]) for i in range(0, len(self.li_one_move), 2)]
            board_move = []
            for index in range(1, len(steps)):
                board_move.append([steps[index - 1], steps[index]])
            self.board_move = board_move

        if self.board_move:
            # steps_move
            if not self.steps_move:
                steps = [self.board_move[0][0]]
                for move in self.board_move:
                    steps.append(move[1])
                self.steps_move = steps

            # li_api_move
            if not self.li_api_move:
                li_api_move = []
                for move in self.board_move:
                    li_api_move.append(self._make_len_2(move[0]) + self._make_len_2(move[1]))
                self.li_api_move = li_api_move

            # li_one_move
            if not self.li_one_move:
                self.li_one_move = "".join(list(map(self._make_len_2, self.steps_move)))

        # Hub related moves

        if self.hub_move:
            # Hub move
            if "-" in self.hub_move:
                self.has_captures = False
                self.hub_position_move = "".join(list(map(self._make_len_2, self.hub_move.split("-"))))
            else:
                self.has_captures = True
                self.hub_position_move = "".join(list(map(self._make_len_2, self.hub_move.split("x"))))
        elif self.hub_position_move:
            # Hub position move
            positions = [self.hub_position_move[i:i + 2] for i in range(0, len(self.hub_position_move), 2)]
            separator = "x" if self.has_captures else "-"
            self.hub_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        if self.hub_to_pdn_pseudolegal:
            # Pseudo-legal PDN move
            self.pdn_position_move = self.hub_position_move[:4]
            separator = "x" if self.has_captures else "-"
            self.pdn_move = self.pdn_position_move[:2] + separator + self.pdn_position_move[2:]

        # PDN related moves

        if self.pdn_move:
            # PDN move
            self.pdn_move = move_from_variant(self.pdn_move, variant=self.variant, notation=self.notation, squares_per_letter=self.squares_per_letter)

            if "-" in self.pdn_move:
                self.has_captures = False
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.pdn_move.split("-"))))
            else:
                self.has_captures = True
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.pdn_move.split("x"))))
        elif self.pdn_position_move:
            # PDN position move
            positions = [self.pdn_position_move[i:i + 2] for i in range(0, len(self.pdn_position_move), 2)]
            separator = "x" if self.has_captures else "-"
            self.pdn_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        if self.pdn_position_move:
            # Get if the move is ambiguous.
            self.ambiguous = not (len(self.pdn_position_move) == 4)



from draughts.engines.dxp import DXPEngine
from draughts.engines.hub import HubEngine
from draughts.engines.checkerboard import CheckerBoardEngine
from typing import Optional, Union, Dict
from draughts.core.move import Move


class Limit:
    def __init__(self, time: Union[int, float, None] = None, inc: Union[int, float, None] = None, depth: Optional[int] = None, nodes: Optional[int] = None, movetime: Union[int, float, None] = None):
        assert time is not None or depth is not None or nodes is not None or movetime is not None
        self.time = time
        self.inc = inc
        self.depth = depth
        self.nodes = nodes
        self.movetime = movetime


class PlayResult:
    def __init__(self, move: Optional[Move] = None, ponder: Optional[Move] = None, info: Optional[Dict] = None, draw_offered: bool = False, resign: bool = False):
        self.move = move
        self.ponder = ponder
        self.info = info
        self.draw_offered = draw_offered
        self.resign = resign


__all__ = ['HubEngine', 'DXPEngine', 'CheckerBoardEngine', 'Limit', 'PlayResult']



from __future__ import annotations
from draughts.core.board import Board
from draughts.core.move import Move
import pickle
from draughts.convert import _algebraic_to_numeric_square, _get_squares
from typing import List, Union, Tuple, Optional

WHITE = 2
BLACK = 1


class Game:

    def __init__(self, variant: str = 'standard', fen: str = 'startpos') -> None:
        self.variant = variant
        if self.variant == 'from position':
            self.variant = 'standard'
        if self.variant == 'american':
            self.variant = 'english'
        elif self.variant == 'frysk':
            self.variant = 'frysk!'
        if fen == 'startpos' or ':' in fen:  # Li fen
            self.initial_fen = self.startpos_to_fen(fen)
            self.initial_hub_fen = self.li_fen_to_hub_fen(self.initial_fen)
            self.board = Board(self.variant, self.initial_hub_fen)
        else:  # Hub fen
            self.initial_hub_fen = fen
            self.board = Board(self.variant, self.initial_hub_fen)
            self.initial_fen = self.get_li_fen()
        self.initial_dxp_fen = self.get_dxp_fen()
        self.last_non_reversible_fen = self.initial_fen

        # moves has every move of a multi-capture as a separate move.
        self.moves = []

        # move_stack contains the moves played but a multi-capture is only considered as one move.
        # move_stack is preferred to moves.
        self.move_stack = []
        self.capture_stack = []

        # _not_added_move and _not_added_capture contain the moves that are part of a multi-capture.
        # that hasn't been completed yet.
        self._not_added_move = []
        self._not_added_capture = []

        # non_reversible_moves contains the moves since the last capture or move of a man.
        # (so it only contains non-capture king moves).
        self.non_reversible_moves = []

        # fens stores the Hub fen of each position to detect threefold repetition.
        self.fens = []
        self.moves_since_last_capture = 0
        self.consecutive_noncapture_king_moves = 0

    def copy(self) -> Game:
        """Copy the board (transfers all data)."""
        # At least 6 times faster than deepcopy.
        return pickle.loads(pickle.dumps(self, -1))

    def copy_fast(self) -> Game:
        """Copy the board (doesn't transfer all the data but is faster)."""
        # More than 10x faster than .copy() but it doesn't transfer all the data.
        game = Game(self.variant, self.get_fen())
        game._not_added_move = self._not_added_move.copy()
        game._not_added_capture = self._not_added_capture.copy()
        return game

    def move(self, move: List[int], return_captured: bool = False) -> Union[Game, Tuple[Game, int]]:
        """Make a move."""
        if move not in self.get_possible_moves():
            raise ValueError('The provided move is not possible')
        turn = self.whose_turn()

        self.board, enemy_position = self.board.push_move(move, len(self.move_stack) + 1, self._not_added_capture)
        self.moves.append(move)

        if self.whose_turn() == turn:
            self._not_added_move.append(move)
            self._not_added_capture.append(enemy_position)
        else:
            captures = self._not_added_capture + [enemy_position]
            if captures[0] is None:
                captures = []
            move_to_add_board = self._not_added_move + [move]
            move_to_add_hub = self.make_len_2(move_to_add_board[0][0]) + self.make_len_2(move_to_add_board[-1][1]) + self.sort_captures(captures)
            move_to_add = Move(board_move=move_to_add_board, hub_position_move=move_to_add_hub, has_captures=bool(captures), hub_to_pdn_pseudolegal=True)
            self.move_stack.append(move_to_add)
            self.capture_stack.append(captures)
            self._not_added_move = []
            self._not_added_capture = []

            piece = self.board.searcher.get_piece_by_position(move[1])
            self.moves_since_last_capture = 0 if self.board.previous_move_was_capture else self.moves_since_last_capture + 1
            if piece.king and not captures:
                self.non_reversible_moves.append(move_to_add)
                self.consecutive_noncapture_king_moves += 1
            else:
                self.last_non_reversible_fen = self.get_li_fen()
                self.non_reversible_moves = []
                self.consecutive_noncapture_king_moves = 0
            self.fens.append(self.get_fen())

        if return_captured:
            return self, enemy_position
        else:
            return self

    def has_player_won(self, player: int = WHITE) -> bool:
        """Get if a given player has won."""
        turn = self.whose_turn()
        opponent_color = WHITE if player == BLACK else BLACK
        if self.variant == 'breakthrough':
            for loc in range(1, self.board.position_count + 1):
                piece = self.board.searcher.get_piece_by_position(loc)
                if piece and piece.player == player and piece.king:
                    # Player wins if they have a king.
                    return True
        elif self.variant == 'antidraughts':
            # Player wins if they have no available move.
            # Can only check if it is the player's turn.
            if turn == player and not self.board.count_movable_player_pieces(player, self._not_added_capture):
                return True
        if self.variant != 'antidraughts':
            # Player wins if the opponent has no available move.
            # Can only check if it is the opponent's turn.
            if turn == opponent_color and not self.board.count_movable_player_pieces(opponent_color, self._not_added_capture):
                return True
        return False

    def is_threefold(self) -> bool:
        """Get if the current position has occurred at least three times."""
        return self.fens and self.fens.count(self.fens[-1]) >= 3

    def is_draw(self) -> bool:
        """Get if the game is a draw."""
        # long_diagonal contains all the squares in the long diagonal of an 8x8 board.
        long_diagonal = [4, 8, 11, 15, 18, 22, 25, 29]
        white_pieces = 0
        black_pieces = 0
        white_kings = 0
        black_kings = 0
        white_piece_in_long_diagonal = False
        black_piece_in_long_diagonal = False
        for loc in range(1, self.board.position_count + 1):
            piece = self.board.searcher.get_piece_by_position(loc)
            if piece is not None:
                if piece.player == WHITE:
                    if loc in long_diagonal:
                        white_piece_in_long_diagonal = True
                    white_pieces += 1
                    if piece.king:
                        white_kings += 1
                else:
                    if loc in long_diagonal:
                        black_piece_in_long_diagonal = True
                    black_pieces += 1
                    if piece.king:
                        black_kings += 1
        if self.variant == 'standard':
            # 25 consecutive non-capture king moves.
            if self.consecutive_noncapture_king_moves >= 50:
                return True
            # 1 king vs 3 pieces (with at least 1 king) and 16 moves made.
            if white_pieces == white_kings == 1 and black_pieces == 3 and black_kings >= 1 and self.moves_since_last_capture >= 32:
                return True
            if black_pieces == black_kings == 1 and white_pieces == 3 and white_kings >= 1 and self.moves_since_last_capture >= 32:
                return True
            # 1 king vs 2 or fewer pieces (with at least 1 king) and 5 moves made.
            if white_pieces == white_kings == 1 and black_pieces <= 2 and black_kings >= 1 and self.moves_since_last_capture >= 10:
                return True
            if black_pieces == black_kings == 1 and white_pieces <= 2 and white_kings >= 1 and self.moves_since_last_capture >= 10:
                return True
            return self.is_threefold()
        elif self.variant in ['frisian', 'frysk!']:
            # 2 kings vs 1 king and 7 moves made.
            if white_pieces == white_kings == 1 and black_pieces == black_kings == 2 and self.moves_since_last_capture >= 14:
                return True
            if white_pieces == white_kings == 2 and black_pieces == black_kings == 1 and self.moves_since_last_capture >= 14:
                return True
            # 1 king vs 1 king and 2 moves made (officially immediately if there can't be a forced capture).
            # but we don't detect if there is a forced capture.
            if white_pieces == white_kings == black_pieces == black_kings == 1 and self.moves_since_last_capture == 4:
                return True
        elif self.variant == 'antidraughts':
            # Only way to draw is threefold.
            return self.is_threefold()
        elif self.variant == 'breakthrough':
            # There is no draw.
            return False
        elif self.variant in ['russian', 'brazilian']:
            # 3 or more kings vs 1 king and 15 moves made.
            if white_pieces == white_kings == 1 and black_pieces == black_kings >= 3 and self.moves_since_last_capture >= 30:
                return True
            if white_pieces == white_kings >= 3 and black_pieces == black_kings == 1 and self.moves_since_last_capture >= 30:
                return True
            # 15 consecutive non-capture king moves.
            if self.consecutive_noncapture_king_moves >= 30:
                return True
            # Same number of kings, same number of pieces, 4 or 5 pieces per side and 30 moves made.
            if white_kings == black_kings >= 1 and white_pieces == black_pieces and white_pieces in [4, 5] and self.moves_since_last_capture >= 60:
                return True
            # Same number of kings, same number of pieces, 6 or 7 pieces per side and 60 moves made.
            if white_kings == black_kings >= 1 and white_pieces == black_pieces and white_pieces in [6, 7] and self.moves_since_last_capture >= 120:
                return True
            # 3 pieces (with at least 1 king) vs 1 king on the long diagonal and 5 moves made.
            if white_pieces == 3 and white_kings >= 1 and black_pieces == black_kings == 1 and not white_piece_in_long_diagonal and black_piece_in_long_diagonal and self.moves_since_last_capture >= 10:
                return True
            if white_pieces == white_kings == 1 and black_pieces == 3 and black_kings >= 1 and white_piece_in_long_diagonal and not black_piece_in_long_diagonal and self.moves_since_last_capture >= 10:
                return True
            # 2 pieces (with at least 1 king) vs 1 king and 5 moves made.
            if white_pieces == 2 and white_kings >= 1 and black_pieces == black_kings == 1 and self.moves_since_last_capture >= 10:
                return True
            if white_pieces == white_kings == 1 and black_pieces == 2 and black_kings >= 1 and self.moves_since_last_capture >= 10:
                return True
            return self.is_threefold()
        elif self.variant in ['english', 'italian']:
            # 40 non-capture king moves.
            if self.consecutive_noncapture_king_moves >= 80:
                return True
            return self.is_threefold()
        elif self.variant == 'turkish':
            # 1 piece (king or not) vs 1 piece (king or not).
            if white_pieces == black_pieces == 1:
                return True
            return self.is_threefold()
        return False

    def is_over(self) -> bool:
        """Get if the game is over."""
        return self.has_player_won(WHITE) or self.has_player_won(BLACK) or self.is_draw()

    def get_winner(self) -> Optional[int]:
        """
        Get the player who won.
        :returns: WHITE if white won, BLACK if black won, 0 if it is a draw, and None if the game hasn't ended.
        """
        if self.has_player_won(WHITE):
            return WHITE
        elif self.has_player_won(BLACK):
            return BLACK
        elif self.is_draw():
            return 0
        return None

    def get_possible_moves(self) -> List[List[int]]:
        """
        Get all possible moves. Doesn't return the whole capture sequence.
        It only returns the first move of a multi-capture, so the moves are pseudo-legal.
        """
        return self.board.get_possible_moves(self._not_added_capture)

    def whose_turn(self) -> int:
        """Get whose turn it is."""
        return self.board.player_turn

    def get_fen(self) -> str:
        """Get the Hub fen of the position."""
        playing = 'W' if self.board.player_turn == WHITE else 'B'
        fen = ''

        for loc in range(1, self.board.position_count + 1):
            piece = self.board.searcher.get_piece_by_position(loc)
            letter = 'e'
            if piece is not None:
                if piece.player == WHITE:
                    letter = 'w'
                else:
                    letter = 'b'
                if piece.king:
                    letter = letter.capitalize()
            fen += letter

        final_fen = playing + fen
        return final_fen

    def get_li_fen(self) -> str:
        """Get the fen of the position."""
        playing = 'W' if self.board.player_turn == WHITE else 'B'
        white_pieces = []
        black_pieces = []

        for loc in range(1, self.board.position_count + 1):
            piece = self.board.searcher.get_piece_by_position(loc)
            if piece is not None:
                letter = str(loc)
                if piece.king:
                    letter = 'K' + letter
                if piece.player == WHITE:
                    white_pieces.append(letter)
                else:
                    black_pieces.append(letter)

        final_fen = f"{playing}:W{','.join(white_pieces)}:B{','.join(black_pieces)}"
        return final_fen

    def get_dxp_fen(self) -> str:
        """Get the DXP fen of the position."""
        fen = ''

        for loc in range(1, self.board.position_count + 1):
            piece = self.board.searcher.get_piece_by_position(loc)
            letter = 'e'
            if piece is not None:
                if piece.player == WHITE:
                    letter = 'w'
                else:
                    letter = 'z'
                if piece.king:
                    letter = letter.capitalize()
            fen += letter

        return fen

    def get_moves(self) -> Tuple[List[List[List[int]]], List[List[Optional[int]]]]:
        """
        Get the moves for a position. It returns the whole multi-capture,
        but it doesn't check the rules of each variant for which multi-capture to choose, so the moves are pseudo-legal.
        Use legal_moves for legal moves.
        """
        turn = self.whose_turn()
        moves = []
        captured_pieces = []
        # get_possible_moves returns only the first jump in a multi-capture sequence,
        # so we use it again after the first jump to check if there are any further moves.
        for move in self.get_possible_moves():
            game_2 = self.copy_fast()
            _, captures = game_2.move(move, return_captured=True)
            if game_2.whose_turn() == turn:
                more_moves, more_captures = game_2.get_moves()
                for semi_move, semi_capture in zip(more_moves, more_captures):
                    moves.append([move] + semi_move)
                    captured_pieces.append([captures] + semi_capture)
            else:
                moves.append([move])
                captured_pieces.append([captures])
        return moves, captured_pieces

    def legal_moves(self) -> Tuple[List[List[List[int]]], List[List[Optional[int]]]]:
        """Get the legal moves for a position."""

        moves, captures = self.get_moves()
        if not moves:
            return moves, captures

        if self.variant in ['frisian', 'frysk!']:
            # Kings are worth 1.5 and men 1. The kings here are worth 1.501 because if we have a choice between 3 men or 2 kings (both are worth 3) we must capture the kings.

            # We have to choose the path that is worth the most.
            king_value = 1.501
            man_value = 1

            values = []
            for capture in captures:
                value = 0
                for position in capture:
                    if position is None:
                        continue
                    piece = self.board.searcher.get_piece_by_position(position)
                    value += king_value if piece.king else man_value
                values.append(value)
            max_value = max(values)
            moves_pseudo_legal = []
            captures_pseudo_legal = []
            for move, capture, value in zip(moves, captures, values):
                if value == max_value:
                    moves_pseudo_legal.append(move)
                    captures_pseudo_legal.append(capture)

            # If a man and a king can play a capture sequence of equal value, it is forced play the capture sequence with the king.
            move_with_king = bool(list(filter(lambda move: self.board.searcher.get_piece_by_position(move[0][0]).king, moves_pseudo_legal)))
            if move_with_king:
                moves_pseudo_legal_2 = []
                captures_pseudo_legal_2 = []
                for move, capture in zip(moves_pseudo_legal, captures_pseudo_legal):
                    if self.board.searcher.get_piece_by_position(move[0][0]).king and capture[0] is not None or capture[0] is None:
                        moves_pseudo_legal_2.append(move)
                        captures_pseudo_legal_2.append(capture)
            else:
                moves_pseudo_legal_2 = moves_pseudo_legal
                captures_pseudo_legal_2 = captures_pseudo_legal

            # The same king can't make more than 3 non-capturing moves in a row, if the player has men left.
            has_man = False
            for loc in range(1, self.board.position_count + 1):
                piece = self.board.searcher.get_piece_by_position(loc)
                if piece and not piece.king and piece.player == self.whose_turn():
                    has_man = True

            if has_man and len(self.move_stack) >= 6:
                last_3_moves = [self.move_stack[-6], self.move_stack[-4], self.move_stack[-2]]
                last_3_moves = list(map(lambda move: move.li_one_move, last_3_moves))
                last_3_moves_same_piece = last_3_moves[0][-2:] == last_3_moves[1][:2] and last_3_moves[1][-2:] == last_3_moves[2][:2]
                was_a_capture = bool(list(filter(bool, [self.capture_stack[-6], self.capture_stack[-4], self.capture_stack[-2]])))
                piece = self.board.searcher.get_piece_by_position(int(last_3_moves[-1][-2:]))
                if piece is None:  # It is None when the piece was captured.
                    is_king = False
                    is_king_for_at_least_3_moves = True
                else:
                    is_king = piece.king
                    is_king_for_at_least_3_moves = len(self.move_stack) - piece.became_king >= 6
                if is_king and last_3_moves_same_piece and not was_a_capture and is_king_for_at_least_3_moves:
                    piece_not_allowed = int(last_3_moves[2][-2:])
                    moves_legal = []
                    captures_legal = []
                    for move, capture in zip(moves_pseudo_legal_2, captures_pseudo_legal_2):
                        if move[0][0] != piece_not_allowed or capture[0] is not None:
                            moves_legal.append(move)
                            captures_legal.append(capture)
                else:
                    moves_legal = moves_pseudo_legal_2
                    captures_legal = captures_pseudo_legal_2
            else:
                moves_legal = moves_pseudo_legal_2
                captures_legal = captures_pseudo_legal_2

        elif self.variant == 'italian':
            # The move that captures the most pieces has to be played.
            max_len_key = max(list(map(len, moves)))
            moves_pseudo_legal = []
            captures_pseudo_legal = []
            for move, capture in zip(moves, captures):
                if len(move) == max_len_key:
                    moves_pseudo_legal.append(move)
                    captures_pseudo_legal.append(capture)

            # If a man and a king can play a capture the same number of pieces, it is forced play the capture sequence with the king.
            move_with_king = bool(list(filter(lambda move: self.board.searcher.get_piece_by_position(move[0][0]).king, moves_pseudo_legal)))
            if move_with_king:
                moves_pseudo_legal_2 = []
                captures_pseudo_legal_2 = []
                for move, capture in zip(moves_pseudo_legal, captures_pseudo_legal):
                    if self.board.searcher.get_piece_by_position(move[0][0]).king and capture[0] is not None or capture[0] is None:
                        moves_pseudo_legal_2.append(move)
                        captures_pseudo_legal_2.append(capture)
            else:
                moves_pseudo_legal_2 = moves_pseudo_legal
                captures_pseudo_legal_2 = captures_pseudo_legal

            # The capture sequence that captures the most kings has to be played.
            max_kings = 0
            for move, capture in zip(moves_pseudo_legal_2, captures_pseudo_legal_2):
                kings = 0
                for piece_loc in capture:
                    if piece_loc is not None and self.board.searcher.get_piece_by_position(piece_loc).king:
                        kings += 1
                max_kings = max(max_kings, kings)
            moves_pseudo_legal_3 = []
            captures_pseudo_legal_3 = []
            for move, capture in zip(moves_pseudo_legal_2, captures_pseudo_legal_2):
                kings = 0
                for piece_loc in capture:
                    if piece_loc is not None and self.board.searcher.get_piece_by_position(piece_loc).king:
                        kings += 1
                if kings == max_kings:
                    moves_pseudo_legal_3.append(move)
                    captures_pseudo_legal_3.append(capture)

            # The capture sequence where the king occurs first has to be played.
            earliest_king = 100
            king_in_capture_sequence = False
            for move, capture in zip(moves_pseudo_legal_3, captures_pseudo_legal_3):
                if capture[0] is not None:
                    for index, piece_loc in enumerate(capture):
                        if self.board.searcher.get_piece_by_position(piece_loc).king:
                            king_in_capture_sequence = True
                            earliest_king = min(earliest_king, index)
                            break
            moves_pseudo_legal_4 = []
            captures_pseudo_legal_4 = []
            if king_in_capture_sequence:
                for move, capture in zip(moves_pseudo_legal_3, captures_pseudo_legal_3):
                    for index, piece_loc in enumerate(capture):
                        if index > earliest_king:
                            break
                        elif self.board.searcher.get_piece_by_position(piece_loc).king:
                            if index == earliest_king:
                                moves_pseudo_legal_4.append(move)
                                captures_pseudo_legal_4.append(capture)
                            break
            else:
                moves_pseudo_legal_4 = moves_pseudo_legal_3
                captures_pseudo_legal_4 = captures_pseudo_legal_3
            moves_legal = moves_pseudo_legal_4
            captures_legal = captures_pseudo_legal_4

        elif self.variant in ['russian', 'english']:
            # No restriction. The player can choose, whichever move they prefer.
            # They only have to complete the multi-capture.
            return moves, captures
        else:
            # Turkish also goes in this category (together with international etc.) because the rule that prohibits a piece from turning 180 degrees is accounted for in get_position_behind_enemy.

            # The move that captures the most pieces has to be played.
            max_len_key = max(list(map(len, moves)))
            moves_legal = []
            captures_legal = []
            for move, capture in zip(moves, captures):
                if len(move) == max_len_key:
                    moves_legal.append(move)
                    captures_legal.append(capture)
        return moves_legal, captures_legal

    def make_len_2(self, move: Union[str, int]) -> str:
        """
        Add a 0 in the front of the square if it is only 1 digit.
        e.g. The move 5 will be turned to 05 but the move 23 will be left the same.
        """
        return f'0{move}' if len(str(move)) == 1 else str(move)

    def push_move(self, move: str) -> None:
        """
        Make a move, given a string.
        e.g. 0523 will mean move the piece at square 5 to square 23.
        """
        self.move([int(move[:2]), int(move[2:4])])

    def sort_captures(self, captures: List[int]) -> str:
        """
        Sort the captures from the smallest number to the highest.
        e.g. [10, 30, 19] will change to '101930'.
        This function exists because hub engines return the captures in alphabetical order
        (e.g. for the move 231201 scan returns 23x01x07x18 instead of 23x01x18x07).
        """
        captures = list(map(self.make_len_2, captures))
        captures.sort()
        captures = ''.join(captures)
        return captures

    def li_fen_to_hub_fen(self, li_fen: str) -> str:
        """Convert a fen to a Hub fen."""
        _, _, squares_per_letter, every_other_square = _get_squares(self.variant)
        fen = ''
        li_fen = li_fen.split(':')
        fen += li_fen[0]
        white_pieces = li_fen[1][1:].split(',')
        black_pieces = li_fen[2][1:].split(',')

        # Fens sometimes contain hyphens to denote that the player has pieces from one square until another.
        # e.g. 5-10 is the same as 5,6,7,8,9,10.
        white_pieces_remove_hyphen = []
        for white_piece in white_pieces:
            if '-' in white_piece:
                start_end = white_piece.split('-')
                add_for_king = ''
                if start_end[0][0] == 'K':
                    add_for_king = 'K'
                    start_end[0] = start_end[0][1:]
                start, end = _algebraic_to_numeric_square(start_end[0], squares_per_letter, every_other_square), _algebraic_to_numeric_square(start_end[1], squares_per_letter, every_other_square)
                for number in range(start, end + 1):
                    white_pieces_remove_hyphen.append(add_for_king + str(number))
            else:
                white_pieces_remove_hyphen.append(white_piece)

        black_pieces_remove_hyphen = []
        for black_piece in black_pieces:
            if '-' in black_piece:
                start_end = black_piece.split('-')
                add_for_king = ''
                if start_end[0][0] == 'K':
                    add_for_king = 'K'
                    start_end[0] = start_end[0][1:]
                start, end = _algebraic_to_numeric_square(start_end[0], squares_per_letter, every_other_square), _algebraic_to_numeric_square(start_end[1], squares_per_letter, every_other_square)
                for number in range(start, end + 1):
                    black_pieces_remove_hyphen.append(add_for_king + str(number))
            else:
                black_pieces_remove_hyphen.append(black_piece)

        position_count = _get_squares(self.variant)[0]

        white_pieces_remove_hyphen = list(map(lambda move: _algebraic_to_numeric_square(move, squares_per_letter) if move[0].lower() != 'k' else move, white_pieces_remove_hyphen))
        black_pieces_remove_hyphen = list(map(lambda move: _algebraic_to_numeric_square(move, squares_per_letter) if move[0].lower() != 'k' else move, black_pieces_remove_hyphen))
        white_pieces_remove_hyphen = list(map(str, white_pieces_remove_hyphen))
        black_pieces_remove_hyphen = list(map(str, black_pieces_remove_hyphen))

        for index in range(1, position_count + 1):
            str_index = str(index)
            if str_index in white_pieces_remove_hyphen:
                fen += 'w'
            elif 'K' + str_index in white_pieces_remove_hyphen:
                fen += 'W'
            elif str_index in black_pieces_remove_hyphen:
                fen += 'b'
            elif 'K' + str_index in black_pieces_remove_hyphen:
                fen += 'B'
            else:
                fen += 'e'
        return fen

    def startpos_to_fen(self, fen: str) -> str:
        """Get the starting fen."""
        if fen != 'startpos':
            return fen
        if self.variant == 'frysk!':
            return 'W:W46,47,48,49,50:B1,2,3,4,5'
        elif self.variant == 'turkish':
            return 'W:W41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56:B9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24'
        elif self.variant in ['brazilian', 'russian', 'english', 'italian']:
            return 'W:W21,22,23,24,25,26,27,28,29,30,31,32:B1,2,3,4,5,6,7,8,9,10,11,12'
        else:
            return 'W:W31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50:B1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20'



import draughts.engines.dxp_communication.dxp_run as dxp
import draughts
import draughts.engine
import subprocess
import os
import signal
import threading
import time
import logging
from importlib import reload
from typing import Optional, Dict, Union, List, Any

logger = logging.getLogger(__name__)


class DXPEngine:
    def __init__(self, command: Union[List[str], str, None] = None, options: Optional[Dict[str, Union[str, int, bool]]] = None, initial_time: int = 0, ENGINE: int = 5) -> None:
        global dxp
        dxp = reload(dxp)
        if options is None:
            options = {}
        self.initial_time = initial_time
        self.max_moves = 0
        self.ip = '127.0.0.1'
        self.port = '27531'
        self.command = command
        self.engine_opened = True  # Whether the engine is already open or pydraughts should open it
        self.wait_to_open_time = 10
        self.ENGINE = ENGINE
        self.info = {}
        self.id = {}
        self.console = dxp.tConsoleHandler
        self.receiver = dxp.tReceiveHandler
        self.game_started = False
        self._last_move = None
        self.exit = False

        for name, value in options.items():
            self.setoption(name, value)

        if not self.engine_opened:
            cwd = os.getcwd()
            cwd = os.path.realpath(os.path.expanduser(cwd))
            if type(command) == str:
                command = [command]
            command = list(filter(bool, command))
            command[0] = os.path.realpath(os.path.expanduser(command[0]))
            command[0] = '"' + command[0] + '"'
            command = ' '.join(command)
            self.command = command
            self.p = self._open_process(command, cwd)
            self.thr = threading.Thread(target=self._recv)
            self.thr.start()

        self.start_time = time.perf_counter_ns()

    def setoption(self, name: str, value: Union[str, int, bool]) -> None:
        """Set a DXP option."""
        if name == 'engine-opened':
            self.engine_opened = value
        elif name == 'ip':
            self.ip = str(value)
        elif name == 'port':
            self.port = str(value)
        elif name == 'wait-to-open-time':
            self.wait_to_open_time = int(value)
        elif name == 'max-moves':
            self.max_moves = 0
        elif name == 'initial-time':
            self.initial_time = 0

    def _open_process(self, command: str, cwd: Optional[str] = None, shell: bool = True, _popen_lock: Any = threading.Lock()) -> subprocess.Popen:
        """Open the engine process."""
        kwargs = {
            "shell": shell,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.STDOUT,
            "stdin": subprocess.PIPE,
            "bufsize": 1,  # Line buffered
            "universal_newlines": True,
        }
        logger.debug(f'command: {command}, cwd: {cwd}')

        if cwd is not None:
            kwargs["cwd"] = cwd

        # Prevent signal propagation from parent process
        try:
            # Windows
            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP
        except AttributeError:
            # Unix
            kwargs["preexec_fn"] = os.setpgrp

        with _popen_lock:  # Work around Python 2 Popen race condition
            return subprocess.Popen(command, **kwargs)

    def kill_process(self) -> None:
        """Kill the engine process."""
        if not self.engine_opened:
            self.exit = True
            try:
                # Windows
                self.p.send_signal(signal.CTRL_BREAK_EVENT)
            except AttributeError:
                # Unix
                os.killpg(self.p.pid, signal.SIGKILL)

            self.p.communicate()

    def _connect(self) -> None:
        """Connect to the engine."""
        if not self.engine_opened:
            wait_time = self.start_time / 1e9 + self.wait_to_open_time - time.perf_counter_ns() / 1e9
            logger.debug(f'wait time: {wait_time}')
            if wait_time > 0:
                time.sleep(wait_time)
        self.console.run_command(f'connect {self.ip} {self.port}')

    def _start(self, board: draughts.Game, time: int) -> None:
        """Start the game."""
        self._connect()
        color = 'B' if board.whose_turn() == draughts.WHITE else 'W'
        time = str(round(time // 60))
        moves = self.max_moves
        self.console.run_command(f'setup {board.initial_fen} {board.variant}')
        self.console.run_command(f'gamereq {color} {time} {moves}')
        accepted = self._recv_accept()
        logger.debug(f'Aceepted: {accepted}')
        self.id["name"] = dxp.current.engineName

    def _recv(self) -> None:
        """Receive a line from the engine, if the engine is opened by pydraughts."""
        # The engine doesn't work otherwise.
        while True:
            try:
                line = self.p.stdout.readline()

                line = line.rstrip()

                if line:
                    logging.debug(f"{self.ENGINE} %s >> %s {self.p.pid} {line}")
            except ValueError as err:
                if self.exit:
                    break
                else:
                    raise err

    def _recv_accept(self) -> bool:
        """Get if the game was accepted."""
        while True:
            if dxp.accepted is not None:
                return dxp.accepted

    def _recv_move(self) -> Optional[List[List[int]]]:
        """Receive the engine move."""
        while True:
            if not dxp.tReceiveHandler.isListening:
                break
            if self._last_move != dxp.last_move:
                self._last_move = dxp.last_move
                logger.debug(f'new last move: {self._last_move}')
                return self._last_move

    def play(self, board: draughts.Game) -> Any:
        """Engine search."""
        if not self.game_started:
            self._start(board, self.initial_time)
            self.game_started = True
        if board.move_stack:
            move = board.move_stack[-1].li_one_move
            move = [move[i:i+2] for i in range(0, len(move), 2)]
            move = '-'.join(move)
            self.console.run_command(f'sm {move}')
        best_move = self._recv_move()
        if best_move:
            best_move = draughts.Move(board, best_move)
        return draughts.engine.PlayResult(best_move, None)

    def quit(self) -> None:
        """Quit the engine."""
        self.console.run_command('gameend 0')



import ctypes
import draughts
from math import ceil


def get_board(board: draughts.Game) -> ctypes.Array:
    """Get a CheckerBoard board (for use in CheckerBoard engines) from a Game() object."""

    row = (ctypes.c_int * 8)
    checkerboard_board = (row * 8)()

    # From CheckerBoard API:
    WHITE = 1
    BLACK = 2
    MAN = 4
    KING = 8
    FREE = 0

    white_starts = board.variant not in ['english']
    flip_board = board.variant not in ['english', 'italian']

    for loc in range(1, board.board.position_count + 1):
        piece = board.board.searcher.get_piece_by_position(loc)
        row = ceil(loc / board.board.width) - 1  # From get_row_from_position

        # Because in english black starts
        if not white_starts:
            row = (board.board.height - 1) - row

        column = (loc - 1) % board.board.width  # From get_column

        # Because:
        # 1. In italian the bottom-left square isn't playable, so in CheckerBoard the board is flipped vertically.
        # 2. In most variants the bottom-left square for the starting side (usually white) is in column a,
        # while in english black starts, so the bottom-left square for the starting side (black) is in row h.
        if flip_board:
            column = board.board.width - 1 - column

        if row % 2 == 0:
            column = (column + 1) * 2 - 2  # To account for the always empty white squares
        else:
            column = (column + 1) * 2 - 1  # To account for the always empty white squares

        number = FREE
        if piece:
            # In Checkerboard black starts first, so the colors are reversed
            if piece.player == draughts.WHITE and not piece.king:
                number = (WHITE if white_starts else BLACK) + MAN
            elif piece.player == draughts.BLACK and not piece.king:
                number = (BLACK if white_starts else WHITE) + MAN
            elif piece.player == draughts.WHITE and piece.king:
                number = (WHITE if white_starts else BLACK) + KING
            elif piece.player == draughts.BLACK and piece.king:
                number = (BLACK if white_starts else WHITE) + KING
        checkerboard_board[column][row] = number

    return checkerboard_board


def from_board(checker_board: ctypes.Array, old_board: draughts.Game) -> str:
    """Get the Hub fen from a CheckerBoard board."""

    # board_numbers = [[square for square in column] for column in checker_board]
    fen = 'B' if old_board.whose_turn() == draughts.WHITE else 'W'  # switch turns

    # From CheckerBoard API:
    WHITE = 1
    BLACK = 2
    MAN = 4
    KING = 8
    FREE = 0

    white_starts = old_board.variant not in ['english']
    flip_board = old_board.variant not in ['english', 'italian']

    for loc in range(1, old_board.board.position_count + 1):
        row = ceil(loc / old_board.board.width) - 1  # From get_row_from_position
        if not white_starts:
            row = (old_board.board.height - 1) - row
        column = (loc - 1) % old_board.board.width  # From get_column

        # Because:
        # 1. In italian the bottom-left square isn't playable, so in CheckerBoard the board is flipped vertically.
        # 2. In most variants the bottom-left square for the starting side (usually white) is in column a,
        # while in english black starts, so the bottom-left square for the starting side (black) is in row h.
        if flip_board:
            column = old_board.board.width - 1 - column

        if row % 2 == 0:
            column = (column + 1) * 2 - 2  # To account for the always empty white squares
        else:
            column = (column + 1) * 2 - 1  # To account for the always empty white squares

        # In Checkerboard black starts first, so the colors are reversed
        if checker_board[column][row] == FREE:
            fen += 'e'
        elif checker_board[column][row] == (BLACK if white_starts else WHITE) + MAN:
            fen += 'b'
        elif checker_board[column][row] == (WHITE if white_starts else BLACK) + MAN:
            fen += 'w'
        elif checker_board[column][row] == (BLACK if white_starts else WHITE) + KING:
            fen += 'B'
        elif checker_board[column][row] == (WHITE if white_starts else BLACK) + KING:
            fen += 'W'
    return fen



# This file is an adaptation of DXC100_draughts_client (https://github.com/akalverboer/DXC100_draughts_client) by akalverboer.

import threading
import logging
from draughts.engines.dxp_communication.dxp_classes import DamExchange, MySocket, GameStatus, DXP_WHITE, DXP_BLACK

last_move = None
accepted = None

logger = logging.getLogger(__name__)


class ConsoleHandler:
    def __init__(self) -> None:
        self.isRunning = False  # not used

    def run_command(self, comm: str) -> None:
        """Send a command to the DXP engine."""

        global current, mySock, lock
        global accepted, last_move
        logger.debug(f'comm {comm}')

        if comm.startswith('q') or comm.startswith('ex'):  # quit/exit
            logger.debug(f"Command terminate program: {comm.strip()}")
            # os._exit(1)  # does no cleanups

        elif comm.startswith('setup'):
            if current.started:
                logger.debug("Game started; setup not allowed")
                return
            if len(comm.split()) == 1:
                # Setup starting position
                logger.debug(f"Command setup starting position: {comm.strip()}")
                current = GameStatus()  # State(Position(board), C.WHITE)
            elif len(comm.split()) == 2:
                # Setup position with fen string (!!! without apostrophes and no spaces !!!)
                _, fen = comm.split(' ', 1)  # strip first word
                logger.debug("Command setup position with FEN string")
                logger.debug(f"FEN: {fen.strip()}")
                color = DXP_BLACK if fen[0] == 'B' else DXP_WHITE
                current = GameStatus(fen=fen, myColor=color)  # State(pos, color)
            elif len(comm.split()) == 3:
                # Setup position with fen string (!!! without apostrophes and no spaces !!!)
                _, fen, variant = comm.split(' ', 2)  # strip first word
                logger.debug("Command setup position with FEN and variant string")
                logger.debug(f"FEN: {fen.strip()} Variant: {variant}")
                color = DXP_BLACK if fen[0] == 'B' else DXP_WHITE
                current = GameStatus(fen=fen, myColor=color, variant=variant)  # State(pos, color)
            else:
                return

        elif comm.startswith('sm'):
            if current.started and current.myColor != current.get_color():
                logger.debug("Move not allowed; server has to move")
                return
            logger.debug(f"Command step move piece: {comm.strip()}")
            steps = comm.strip().split()[1].split('-')
            steps = list(map(int, steps))

            lock.acquire()  # LOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCK
            if current.started and current.myColor == current.get_color():
                timeSpend = 0  # time spend for this move (future)

                board = []
                for index in range(1, len(steps)):
                    board.append([steps[index - 1], steps[index]])
                all_moves, all_captures = current.pos.legal_moves()
                captures = all_captures[all_moves.index(board)]
                if captures[0] is None:
                    captures = []
                msg = dxp.msg_move(steps, captures, timeSpend)
                logger.debug(f'MOVE: {board}')

                try:
                    mySock.send(msg)
                    logger.debug(f"snd MOVE: {msg}")
                except Exception as err:
                    logger.debug(f"Error sending move: {err}")
                    return

                for move in board:
                    current.pos.move(move)
            lock.release()  # LOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCK

        elif comm.startswith('conn'):
            if mySock.sock is not None:
                logger.debug("Already connected")
                return
            if current.started:
                logger.debug("Game marked as started. First exit to start a new game.")
                return
            host, port = '127.0.0.1', 27531  # default
            if len(comm.split()) == 2:
                _, host = comm.split()
            if len(comm.split()) == 3:
                _, host, port = comm.split()
            logger.debug(f"Command make connection with host {host} port {port}")
            try:
                mySock.open()
                mySock.connect(host, int(port))  # with timeout
                logger.debug(f"Successfully connected to remote host {host}, port {port}")
            except Exception as err:
                mySock.sock = None
                logger.debug(f"Error trying to connect: {err}")
                return

            if mySock.sock is not None:  # connected
                if not tReceiveHandler.isListening:
                    tReceiveHandler.start()

        elif comm.startswith('chat'):
            if len(comm.split()) == 1:
                return
            if len(comm.split()) > 1:
                _, txt = comm.split(' ', 1)  # strip first word
                txt = txt.strip()  # trim whitespace
                logger.debug(f"Command send chat message: {comm.strip()}")
                msg = dxp.msg_chat(txt)
                try:
                    mySock.send(msg)
                    logger.debug(f"snd CHAT: {msg}")
                except Exception as err:
                    logger.debug(f"Error sending chat message: {err}")
                    return

        elif comm.startswith('gamereq'):
            if current.started:
                logger.debug("Game already started; gamereq not allowed")
                return
            logger.debug(f"Command request new game: {comm.strip()}")
            last_move = None
            accepted = None
            myColor = "W"  # default
            gameTime = "120"  # default
            numMoves = "50"  # default
            if len(comm.split()) == 2:
                _, myColor = comm.split()
            if len(comm.split()) == 3:
                _, myColor, gameTime = comm.split()
            if len(comm.split()) == 4:
                _, myColor, gameTime, numMoves = comm.split()

            myColor = DXP_WHITE if myColor.upper().startswith('W') else DXP_BLACK
            current.myColor = myColor  # 0 or 1
            current.gameTime = gameTime
            current.numMoves = numMoves
            msg = dxp.msg_gamereq(myColor, gameTime, numMoves, current.pos, current.get_color())
            try:
                mySock.send(msg)
                logger.debug(f"snd GAMEREQ: {msg}")
            except Exception as err:
                logger.debug(f"Error sending game request: {err}")

        elif comm.startswith('gameend'):
            if not current.started:
                logger.debug("Game already finished; gameend not allowed")
                return
            logger.debug(f"Command finish game: {comm.strip()}")
            if current.started and current.myColor != current.get_color():
                logger.debug("Message gameend not allowed; wait until your turn")
                return
            if len(comm.split()) == 2:
                _, reason = comm.split()
            else:
                reason = "0"
            msg = dxp.msg_gameend(reason)

            try:
                mySock.send(msg)
                logger.debug(f"snd GAMEEND: {msg}")
            except Exception as err:
                logger.debug(f"Error sending gameend message: {err}")

        elif comm.startswith('backreq'):
            if not current.started:
                logger.debug("Game not started; backreq not allowed")
                return
            logger.debug("Not yet supported")

        else:
            logger.debug(f"Command unknown: {comm.strip()}")
            logger.debug(f"Unknown command, type h for help: {comm.strip()}")


class ReceiveHandler(threading.Thread):
    # Subslass of Thread to handle incoming messages from client.

    def __init__(self) -> None:
        threading.Thread.__init__(self)
        self.isListening = False

    def run(self) -> None:
        """Start the receiver. It receives the moves from the DXP engine."""
        # Handling incoming messages from server.
        # Excutes when thread started. Overriding python threading.Thread.run()

        logger.debug("ReceiveHandler started")
        global current, mySock, lock
        global accepted, last_move
        self.isListening = True
        logger.debug("DXP Client starts listening")
        while True:
            try:
                message = mySock.receive()  # wait for message
            except Exception as err:
                logger.debug(f"Error {err}")
                break

            lock.acquire()  # LOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCK
            message = message[0:127]  # DXP max length
            dxpData = dxp.parse(message)
            if dxpData["type"] == "C":
                logger.debug(f"rcv CHAT: {message}")
                logger.debug(f"\nChat message: {dxpData['text']}")

            elif dxpData["type"] == "A":
                logger.debug(f"rcv GAMEACC: {message}")
                if dxpData["accCode"] == "0":
                    current.started = True
                    current.myColor = current.myColor  # as requested
                    current.engineName = dxpData["engineName"]
                    logger.debug(f"\nGame request accepted by {dxpData['engineName']}")
                    accepted = True
                else:
                    current.started = False
                    logger.debug(f"\nGame request NOT accepted by {dxpData['engineName']} Reason: {dxpData['accCode']}")
                    accepted = False

            elif dxpData["type"] == "E":
                logger.debug(f"rcv GAMEEND: {message}")
                logger.debug(f"\nRequest end of game accepted. Reason: {dxpData['reason']} Stop: {dxpData['stop']}")
                # Confirm game end by sending message back (if not sent by me)
                if current.started:
                    current.started = False
                    current.result = dxpData["reason"]
                    msg = dxp.msg_gameend(dxpData["reason"])
                    mySock.send(msg)
                    logger.debug(f"snd GAMEEND: {msg}")

            elif dxpData["type"] == "M":
                logger.debug(f"rcv MOVE: {message}")
                steps = [dxpData['from'], dxpData['to']]
                nsteps = list(map(int, steps))
                ntakes = list(map(int, dxpData['captures']))
                board = None
                all_moves, all_captures = current.pos.legal_moves()
                if not ntakes:
                    ntakes = [None]
                for move, capture in zip(all_moves, all_captures):
                    if move[0][0] == nsteps[0] and move[-1][1] == nsteps[-1] and sorted(ntakes) == sorted(capture):
                        board = move

                if board is not None:
                    last_move = board
                    logger.debug(f"\nMove received: {board}")
                    for move in board:
                        current.pos.move(move)
                else:
                    logger.debug(f"Error: received move is illegal [{message}]")

            elif dxpData["type"] == "B":
                # For the time being do not confirm request from server: send message back.
                logger.debug(f"rcv BACKREQ: {message}")
                accCode = "1"  # 0: BACK YES; 1: BACK NO; 2: CONTINUE
                msg = dxp.msg_backacc(accCode)
                mySock.send(msg)
                logger.debug(f"snd BACKACC: {msg}")

            elif dxpData["type"] == "K":
                # Answer to my request to move back
                logger.debug(f"rcv BACKACC: {message}")
                logger.debug(f"rcv BACKACC: {message}")  # TEST
                accCode = dxpData['accCode']
                if accCode == "0":
                    # Actions to go back in history as specified in my request
                    raise NotImplementedError("Moving back is not implemented")

            else:
                logger.debug(f"rcv UNKNOWN: {message}")
                logger.debug(f"\nrcv Unknown message: {message}")

            lock.release()  # LOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCKLOCK

        self.isListening = False
        logger.error("Listening stopped; connection broken")
        logger.debug("Connection broken; receiveHandler stopped. ")
        logger.debug("Save your data and exit program to start again. ")
        return None


dxp = DamExchange()  # global, singleton
mySock = MySocket()  # global, singleton
current = GameStatus()  # global; use default parms
lock = threading.Lock()  # global

# use 1 thread to listen to incoming messages
tConsoleHandler = ConsoleHandler()  # Thread subclass instance
tReceiveHandler = ReceiveHandler()  # Thread subclass instance. Start when connected



# 11 english, 11 italian, 4 move english and 5 move english ballots are from Ed Gilbert (http://edgilbert.org/Checkers/KingsRow.htm).
import random
import json
import os
from typing import Tuple, List, Dict


class Ballots:
    def __init__(self, variant: str, moves: int = 3, eleven_pieces: bool = False, basic_positions: bool = False, include_lost_games: bool = False) -> None:
        self.variant = variant
        self.moves = moves
        self.eleven_pieces = eleven_pieces
        self.basic_positions = basic_positions
        self.include_lost_games = include_lost_games
        self.filename = self._find_file()
        self.positions, self.keys = self.open_file()
        self.keys_to_use = self.keys.copy()

    def _find_file(self) -> str:
        """Get the filename of the ballots."""
        if self.variant == 'italian':
            return '11_italian.json'
        if self.variant == 'english':
            if self.eleven_pieces:
                return '11_english.json'
            if self.moves == 2:
                return '2move_english.json'
            if self.moves == 4:
                return '4move_english.json'
            if self.moves == 5:
                return '5move_english.json'
        if self.basic_positions:
            return '150russian_and_brazilian.json'
        if self.variant == 'russian':
            return 'russian.json'
        if self.variant == 'brazilian':
            return 'brazilian.json'
        return '3move_english.json'

    def open_file(self) -> Tuple[Dict[str, str], List[str]]:
        """Open the ballot file."""
        filepath = os.path.join(os.path.dirname(__file__), 'ballot_files', self.filename)
        with open(filepath) as file:
            data = json.load(file)
        keys = data['standard'] + (data.get('lost', []) if self.include_lost_games else [])
        return data['all'], keys

    def get_ballot(self) -> str:
        """Get one ballot."""
        if not self.keys_to_use:
            self.keys_to_use = self.keys.copy()
        key = self.keys_to_use.pop(random.randint(0, len(self.keys_to_use) - 1))
        return self.positions[key]



from __future__ import annotations
from draughts.core.board_searcher import BoardSearcher
from draughts.core.board_initializer import BoardInitializer
from functools import reduce
import pickle
from typing import Optional, List, Tuple, Any

WHITE = 2
BLACK = 1


class Board:

    def __init__(self, variant: str = 'standard', fen: str = 'startpos') -> None:
        if fen != 'startpos':
            self.player_turn = WHITE if fen[0].lower() == 'w' else BLACK
        else:
            self.player_turn = WHITE

        if variant in ['brazilian', 'russian', 'english', 'italian']:
            self.width = 4
            self.height = 8
        elif variant == 'turkish':
            self.width = 8
            self.height = 8
        else:
            self.width = 5
            self.height = 10

        if variant == 'frysk!':
            self.rows_per_user_with_pieces = 1
        elif variant == 'turkish':
            self.rows_per_user_with_pieces = 2
        elif variant in ['brazilian', 'russian', 'english', 'italian']:
            self.rows_per_user_with_pieces = 3
        else:
            self.rows_per_user_with_pieces = 4

        self.position_count = self.width * self.height
        self.position_layout = {}
        self.piece_requiring_further_capture_moves = None
        self.previous_move_was_capture = False
        self.variant = variant
        self.fen = fen
        self.searcher = BoardSearcher()
        BoardInitializer(self, self.fen).initialize()

        self.pieces_promote_and_stop_capturing = self.variant in ['english', 'italian']
        self.pieces_promote_and_continue_capturing = self.variant in ['russian']

    def count_movable_player_pieces(self, player_number: int = 1, captures: Optional[List[int]] = None) -> int:
        """Count the pieces of one player that can be moved."""
        if captures is None:
            captures = []
        return reduce((lambda count, piece: count + (1 if piece.is_movable(captures) else 0)), self.searcher.get_pieces_by_player(player_number), 0)

    def get_possible_moves(self, captures: List[int]) -> List[List[int]]:
        """Get all possible moves."""
        capture_moves = self.get_possible_capture_moves(captures)

        return capture_moves if capture_moves else self.get_possible_positional_moves()

    def get_possible_capture_moves(self, captures: List[int]) -> List[List[int]]:
        """Get all possible capture moves (not positional moves)."""
        return reduce((lambda moves, piece: moves + piece.get_possible_capture_moves(captures)), self.searcher.get_pieces_in_play(), [])

    def get_possible_positional_moves(self) -> List[List[int]]:
        """Get all possible positional moves (not capture moves)."""
        return reduce((lambda moves, piece: moves + piece.get_possible_positional_moves()), self.searcher.get_pieces_in_play(), [])

    def position_is_open(self, position: int) -> bool:
        """Get if the position is open (a piece is not in the given square)."""
        return not self.searcher.get_piece_by_position(position)

    def create_new_board_from_move(self, move: List[int], move_number: int, captures: List[int]) -> Tuple[Board, Optional[int]]:
        """Create a new board and play the move given."""
        new_board = pickle.loads(pickle.dumps(self, -1))  # A lot faster that deepcopy.
        enemy_position = None

        if move in self.get_possible_capture_moves(captures):
            enemy_position = new_board.perform_capture_move(move, move_number, captures)
        else:
            new_board.perform_positional_move(move, move_number)

        return new_board, enemy_position

    def push_move(self, move: List[int], move_number: int, captures: List[int]) -> Tuple[Board, Optional[int]]:
        """Play the move given without creating a new board."""
        # It takes 40% less time than create_new_board_from_move (60% faster).
        enemy_position = None

        if move in self.get_possible_capture_moves(captures):
            enemy_position = self.perform_capture_move(move, move_number, captures)
        else:
            self.perform_positional_move(move, move_number)

        return self, enemy_position

    def perform_capture_move(self, move: List[int], move_number: int, captures: List[int]) -> Optional[int]:
        """Make a capture move."""
        self.previous_move_was_capture = True
        piece = self.searcher.get_piece_by_position(move[0])
        originally_was_king = piece.king
        enemy_piece = piece.capture_move_enemies[move[1]]
        enemy_position = enemy_piece.position
        enemy_piece.capture()
        self.move_piece(move, move_number)
        if not originally_was_king and piece.king and self.pieces_promote_and_stop_capturing:
            further_capture_moves_for_piece = []
        elif not originally_was_king and not self.pieces_promote_and_continue_capturing:
            was_king = piece.king
            piece.king = False
            further_capture_moves_for_piece = [capture_move for capture_move in self.get_possible_capture_moves(captures + [enemy_position]) if move[1] == capture_move[0]]
            if not further_capture_moves_for_piece and was_king:
                piece.king = True
        else:
            further_capture_moves_for_piece = [capture_move for capture_move in self.get_possible_capture_moves(captures + [enemy_position]) if move[1] == capture_move[0]]

        if further_capture_moves_for_piece:
            self.piece_requiring_further_capture_moves = self.searcher.get_piece_by_position(move[1])
        else:
            self.piece_requiring_further_capture_moves = None
            self.switch_turn()
        return enemy_position

    def perform_positional_move(self, move: List[int], move_number: int) -> None:
        """Make a positional move."""
        self.previous_move_was_capture = False
        self.move_piece(move, move_number)
        self.switch_turn()

    def switch_turn(self) -> None:
        """Switch the turn."""
        self.player_turn = BLACK if self.player_turn == WHITE else WHITE

    def move_piece(self, move: List[int], move_number: int) -> None:
        """Move a piece."""
        self.searcher.get_piece_by_position(move[0]).move(move[1], move_number)
        self.pieces = sorted(self.pieces, key=lambda piece: piece.position if piece.position else 0)

    def is_valid_row_and_column(self, row: int, column: int) -> bool:
        """Get if the given row and column is inside the board."""
        if row < 0 or row >= self.height:
            return False

        if column < 0 or column >= self.width:
            return False

        return True

    def __setattr__(self, name: str, value: Any) -> None:
        super(Board, self).__setattr__(name, value)

        if name == 'pieces':
            [piece.reset_for_new_board() for piece in self.pieces]

            self.searcher.build(self)



# This file is an adaptation of fishnet (https://github.com/lichess-org/fishnet/tree/ebd2a5e16d37135509cbfbff9998e0b798866ef5).

import subprocess
import os
import signal
import logging
import threading
import draughts
import draughts.engine
import re
import math
from typing import Union, Optional, List, Tuple, Any, Dict, Set

logger = logging.getLogger(__name__)


class HubEngine:
    def __init__(self, command: Union[List[str], str], cwd: str = None, ENGINE: int = 5) -> None:
        self.ENGINE = ENGINE
        self.info = {}
        self.id = {}
        self.options = set()
        self.variants = set()
        cwd = cwd or os.getcwd()
        cwd = os.path.realpath(os.path.expanduser(cwd))
        if type(command) == str:
            command = [command]
        command = list(filter(bool, command))
        command[0] = os.path.realpath(os.path.expanduser(command[0]))
        command[0] = '"' + command[0] + '"'
        command = ' '.join(command)
        self.p = self._open_process(command, cwd)
        self._last_sent = ""
        self.hub()

    def _open_process(self, command: str, cwd: Optional[str] = None, shell: Optional[bool] = True, _popen_lock: Any = threading.Lock()) -> subprocess.Popen:
        """Open the engine process."""
        kwargs = {
            "shell": shell,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.STDOUT,
            "stdin": subprocess.PIPE,
            "bufsize": 1,  # Line buffered
            "universal_newlines": True,
        }

        if cwd is not None:
            kwargs["cwd"] = cwd

        # Prevent signal propagation from parent process
        try:
            # Windows
            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP
        except AttributeError:
            # Unix
            kwargs["preexec_fn"] = os.setpgrp

        with _popen_lock:  # Work around Python 2 Popen race condition
            return subprocess.Popen(command, **kwargs)

    def kill_process(self) -> None:
        """Kill the engine process."""
        try:
            # Windows
            self.p.send_signal(signal.CTRL_BREAK_EVENT)
        except AttributeError:
            # Unix
            os.killpg(self.p.pid, signal.SIGKILL)

        self.p.communicate()

    def send(self, line: str) -> None:
        """Send a command to the engine."""
        if line == "ponder-hit":
            while self._last_sent != "go ponder":
                pass
        logger.debug(f"{self.ENGINE} %s << %s {self.p.pid} {line}")

        self.p.stdin.write(line + "\n")
        self.p.stdin.flush()
        self._last_sent = line

    def recv(self) -> str:
        """Receive a line from the engine."""
        while True:
            line = self.p.stdout.readline()
            if line == "":
                raise EOFError()

            line = line.rstrip()

            logger.debug(f"{self.ENGINE} %s >> %s {self.p.pid} {line}")

            if line:
                return line

    def recv_hub(self) -> List[str]:
        """Receive a line from the engine and split at the first space."""
        command_and_args = self.recv().split(None, 1)
        if len(command_and_args) == 1:
            return [command_and_args[0], ""]
        elif len(command_and_args) == 2:
            return command_and_args

    def hub(self) -> Tuple[Dict[str, str], Set[str], Set[str]]:
        """Send the hub command to an engine."""
        self.send("hub")

        engine_info = {}
        options = set()
        variants = set()

        while True:
            command, arg = self.recv_hub()

            if command == "wait":
                return engine_info, options, variants
            elif command == "id":
                args = re.split(r' +(?![^"]*"(?:(?:[^"]*"){2})*[^"]*$)', arg)
                args = list(map(lambda item: item.split("="), args))
                for key, value in args:
                    engine_info[key] = value
            elif command == "param":
                args = re.split(r' +(?![^"]*"(?:(?:[^"]*"){2})*[^"]*$)', arg)
                args = list(map(lambda item: item.split("="), args))
                is_variant = False
                for key, value in args:
                    if key == "name":
                        options.add(value)
                        if value == "variant":
                            is_variant = True
                    if key == "values" and is_variant:
                        if value.startswith('"') and value.endswith('"'):
                            value = value[1:-1]
                        for variant in value.split():
                            variants.add(variant)
            else:
                logger.warning("Unexpected engine response to hub: %s %s", command, arg)
            self.options = options
            self.variants = variants
            self.id = engine_info

    def init(self) -> None:
        """Initialize the engine."""
        self.send("init")
        while True:
            command, arg = self.recv_hub()
            if command == "ready":
                break
            elif command == "init":
                pass
            else:
                logger.warning("Unexpected engine response to init: %s %s", command, arg)

    def ping(self) -> None:
        """Send the engine ping. They should reply with pong."""
        self.send("ping")
        while True:
            command, arg = self.recv_hub()
            if command == "pong":
                break
            else:
                logger.warning("Unexpected engine response to ping: %s %s", command, arg)

    def setoption(self, name: str, value: Union[str, bool]) -> None:
        """Set an engine option."""
        if value is True:
            value = "true"
        elif value is False:
            value = "false"
        elif value is None:
            value = "none"

        if name == 'variant' and self.variants or name != 'variant':
            self.send("set-param name=%s value=%s" % (name, value))

    def go(self, fen: str, moves: Optional[str] = None, my_time: Union[int, float, None] = None, inc: Union[int, float, None] = None, moves_left: Optional[int] = None, movetime: Union[int, float, None] = None, depth: Optional[int] = None, nodes: Optional[int] = None, ponder: Optional[bool] = False) -> Tuple[str, Optional[str]]:
        """Send the engine a go command."""
        if moves:
            self.send(f'pos pos={fen} moves="{moves}"')
        else:
            self.send(f'pos pos={fen}')

        if my_time is not None and inc and moves_left:
            my_time -= inc  # Hub engines first add the increment
            self.send(f'level moves={moves_left} time={my_time} inc={inc}')
        elif my_time is not None and inc:
            my_time -= inc  # Hub engines first add the increment
            self.send(f'level time={my_time} inc={inc}')
        elif my_time is not None and moves_left:
            self.send(f'level moves={moves_left} time={my_time}')
        elif my_time is not None:
            self.send(f'level time={my_time}')
        elif movetime is not None:
            self.send(f'level move-time={movetime}')
        elif depth is not None:
            self.send(f'level depth={depth}')
        elif nodes is not None:
            self.send(f'level nodes={nodes}')

        if ponder:
            self.send('go ponder')
        else:
            self.send('go think')

        self.info = {}
        while True:
            command, arg = self.recv_hub()
            if command == "done":
                args = arg.split()
                pondermove = None
                args = list(map(lambda item: item.split("="), args))
                bestmove = args[0][1]
                if len(args) == 2:
                    pondermove = args[1][1]
                return bestmove, pondermove
            elif command == "info":
                args = re.split(r' +(?![^"]*"(?:(?:[^"]*"){2})*[^"]*$)', arg)
                args = list(map(lambda item: item.split("="), args))
                for key, value in args:
                    if key in ["depth", "nodes"]:
                        value = int(value)
                    elif key in ["mean-depth", "time", "nps"]:
                        value = float(value)
                    elif key == "score":
                        score = int(float(value) * 100)
                        mate = None
                        if score > 9000:
                            mate = 10000 - score
                        elif score < -9000:
                            mate = -10000 - score
                        if mate:
                            value = {"win": math.ceil(mate / 2)}
                        else:
                            value = {"cp": score}
                    self.info[key] = value
            else:
                logger.warning("Unexpected engine response to go: %s %s", command, arg)

    def stop(self) -> None:
        """Stop the engine from searching."""
        self.send("stop")

    def ponderhit(self) -> None:
        """Send ponder-hit to the engine."""
        self.send("ponder-hit")

    def quit(self) -> None:
        """Quit the engine."""
        self.send("quit")

    def play(self, board: draughts.Game, time_limit: Any, ponder: bool) -> Any:
        """Engine search."""
        time = time_limit.time
        inc = time_limit.inc
        depth = time_limit.depth
        nodes = time_limit.nodes
        movetime = time_limit.movetime
        hub_moves = board.move_stack
        hub_moves = list(map(lambda move: move.hub_move, hub_moves))
        bestmove, pondermove = self.go(board.initial_hub_fen, moves=' '.join(hub_moves), my_time=time, inc=inc, depth=depth, nodes=nodes, movetime=movetime, ponder=ponder)

        ponder_move = None
        ponder_board = board.copy()
        best_move = draughts.Move(ponder_board, hub_move=bestmove)
        if pondermove:
            for move in best_move.board_move:
                ponder_board.move(move)
            ponder_move = draughts.Move(ponder_board, hub_move=pondermove)

        return draughts.engine.PlayResult(best_move, ponder_move, self.info)



from draughts.core.piece import Piece
from typing import Any

WHITE = 2
BLACK = 1


class BoardInitializer:

    def __init__(self, board: Any, fen: str = 'startpos') -> None:
        self.board = board
        self.fen = fen

    def initialize(self) -> None:
        """Initialize the board."""
        self.build_position_layout()
        self.set_starting_pieces()

    def build_position_layout(self) -> None:
        """Build the position layout."""
        self.board.position_layout = {}
        position = 1

        for row in range(self.board.height):
            self.board.position_layout[row] = {}

            for column in range(self.board.width):
                self.board.position_layout[row][column] = position
                position += 1

    def set_starting_pieces(self) -> None:
        """Create the pieces."""
        pieces = []
        if self.fen != 'startpos':  # Hub fen
            # starting = self.fen[0]
            board = self.fen[1:]
            for index, position in enumerate(board):
                piece = None
                if position.lower() == 'w':
                    # Index + 1 because enumerate returns 0-49 while the board takes 1-50.
                    piece = self.create_piece(2, index + 1)
                elif position.lower() == 'b':
                    piece = self.create_piece(1, index + 1)
                if position == 'W' or position == 'B':
                    piece.king = True
                if piece:
                    pieces.append(piece)
        else:  # Not used, but it isn't removed, because it may be needed in the future
            starting_piece_count = self.board.width * self.board.rows_per_user_with_pieces
            player_starting_positions = {
                1: list(range(1, starting_piece_count + 1)),
                2: list(range(self.board.position_count - starting_piece_count + 1, self.board.position_count + 1))
            }

            for key, row in self.board.position_layout.items():
                for key, position in row.items():
                    player_number = 1 if position in player_starting_positions[1] else 2 if position in player_starting_positions[2] else None

                    if player_number:
                        pieces.append(self.create_piece(player_number, position))

        self.board.pieces = pieces

    def create_piece(self, player_number: int, position: int) -> Piece:
        """Create a piece."""
        piece = Piece(variant=self.board.variant)
        piece.player = player_number
        piece.position = position
        piece.board = self.board

        return piece



from __future__ import annotations
from math import ceil
from typing import List

WHITE = 2
BLACK = 1


class Piece:

    def __init__(self, variant: str = 'standard') -> None:
        self.player = None
        self.king = False
        self.captured = False
        self.position = None
        self.board = None
        self.became_king = -100
        self.capture_move_enemies = {}
        self.variant = variant
        self.reset_for_new_board()

        self.diagonal_moves = self.variant not in ['turkish']
        self.orthogonal_moves = self.variant in ['turkish']
        self.orthogonal_captures = self.variant in ['frisian', 'frysk!', 'turkish']
        # In most draughts variants only the dark squares are playable.
        self.half_of_the_squares_are_playable = self.variant not in ['turkish']
        # The bottom left square isn't a playing square in italian draughts.
        self.bottom_left_square_is_playable = self.variant in ['italian']
        # Men can't capture kings in italian draughts.
        self.man_can_capture_king = self.variant not in ['italian']
        # There are three possible values for squares_per_row (4, 5, 8), which is why two lines are needed.
        self.squares_per_row = 4 if self.variant in ['russian', 'brazilian', 'english', 'italian'] else 5
        self.squares_per_row = 8 if self.variant in ['turkish'] else self.squares_per_row
        self.kings_can_move_more_than_one_square = self.variant not in ['english', 'italian']
        self.men_can_capture_backwards = self.variant not in ['english', 'italian', 'turkish']
        self.kings_can_jump_over_an_already_captured_piece = self.variant in ['turkish']
        self.kings_can_turn_180_degrees_in_multicapture = False  # No variant supports it for now.

    def reset_for_new_board(self) -> None:
        """Reset possible moves to None."""
        self.possible_capture_moves = None
        self.possible_positional_moves = None

    def get_square(self, row: int, column: int) -> int:
        """Get the square given the row and column."""
        return self.board.position_layout.get(row, {}).get(column)

    def is_movable(self, captures: List[int]) -> bool:
        """Get if the piece can move."""
        return bool((self.get_possible_capture_moves(captures) or self.get_possible_positional_moves()) and not self.captured)

    def capture(self) -> None:
        """Flag the piece as captured."""
        self.captured = True
        self.position = None

    def move(self, new_position: int, move_number: int) -> None:
        """Move the piece to a new position."""
        self.position = new_position
        was_king = self.king
        self.king = self.king or self.is_on_enemy_home_row()
        if self.king != was_king:
            self.became_king = move_number

    def get_possible_capture_moves(self, captures: List[int]) -> List[List[int]]:
        """Get all possible capture moves (not positional moves) for this piece."""
        if self.possible_capture_moves is None:
            self.possible_capture_moves = self.build_possible_capture_moves(captures)

        return self.possible_capture_moves

    def build_possible_capture_moves(self, captures: List[int]) -> List[List[int]]:
        """Build all possible capture moves (not positional moves) for this piece."""
        adjacent_enemy_positions = list(filter((lambda position: position in self.board.searcher.get_positions_by_player(self.other_player)), self.get_adjacent_positions(capture=True)))
        capture_move_positions = []

        for enemy_position in adjacent_enemy_positions:
            enemy_piece = self.board.searcher.get_piece_by_position(enemy_position)

            if not self.man_can_capture_king and not self.king and enemy_piece.king:
                continue

            positions_behind_enemy = self.get_position_behind_enemy(enemy_piece, captures)
            for position_behind_enemy in positions_behind_enemy:

                if (position_behind_enemy is not None) and self.board.position_is_open(position_behind_enemy):
                    capture_move_positions.append(position_behind_enemy)
                    self.capture_move_enemies[position_behind_enemy] = enemy_piece

        return self.create_moves_from_new_positions(capture_move_positions)

    def get_diagonal_one_square_behind_enemy(self, enemy_piece: Piece) -> List[int]:
        """
        Get the diagonal square directly behind the enemy piece if the piece can move only one square
        (usually men, but in some variants kings can also only move one square).
        """
        # Diagonal captures
        current_row = self.get_row()
        current_column = self.get_column()
        enemy_column = enemy_piece.get_column()
        enemy_row = enemy_piece.get_row()

        # Orthogonal, not diagonal
        if current_row == enemy_row or current_column == enemy_column and (current_row - enemy_row) % 2 == 0:
            return []

        column_adjustment = -1 if current_row % 2 == int(self.bottom_left_square_is_playable) else 1
        column_behind_enemy = current_column + column_adjustment if current_column == enemy_column else enemy_column
        row_behind_enemy = enemy_row + (enemy_row - current_row)

        return [self.get_square(row_behind_enemy, column_behind_enemy)]

    def get_orthogonal_one_square_behind_enemy(self, enemy_piece: Piece) -> List[int]:
        """
        Get the orthogonal square directly behind the enemy piece if the piece can move only one square
        (usually men, but in some variants kings can also only move one square).
        """
        # Orthogonal captures
        current_row = self.get_row()
        current_column = self.get_column()
        enemy_column = enemy_piece.get_column()
        enemy_row = enemy_piece.get_row()

        row_difference = current_row - enemy_row
        column_difference = current_column - enemy_column

        # Check if the pieces are on the same row or column

        # If half_of_the_squares_are_playable the square in the front is not playable
        # (e.g. a1 is playable but a2 isn't, so the square directly in front of it is a3)
        if (column_difference == 0 or row_difference == 0) and (row_difference % 2 == 0 if self.half_of_the_squares_are_playable else True):
            next_row = enemy_row - row_difference
            next_column = enemy_column - column_difference
            return [self.get_square(next_row, next_column)]
        return []

    def get_diagonal_multiple_squares_behind_enemy(self, enemy_piece: Piece, captures: List[int]) -> List[int]:
        """Get the diagonal squares behind the enemy piece if the piece can move more than one square (kings)."""
        positions = []
        current_row = self.get_row()
        current_column = self.get_column()
        enemy_column = enemy_piece.get_column()
        enemy_row = enemy_piece.get_row()

        # Orthogonal, not diagonal
        if current_row == enemy_row or current_column == enemy_column and (current_row - enemy_row) % 2 == 0:
            return []

        adjacent_positions = self.get_adjacent_positions(capture=True)
        down_direction = enemy_row > current_row
        left_direction = enemy_column < current_column if current_row % 2 == 1 else enemy_column <= current_column

        legal_adjacent_positions = []
        was_king = enemy_piece.king
        enemy_piece.king = True

        for position in enemy_piece.get_adjacent_positions(capture=True):
            column = (position - 1) % self.board.width
            row = self.get_row_from_position(position)
            down_direction_possible = row > enemy_row
            left_direction_possible = column < enemy_column if enemy_row % 2 == 1 else column <= enemy_column
            # Checks if the captured piece square is between the starting square and the landing square
            # For example, if the captured piece square is to the left and down of the starting square, we check if the landing square is to the left and down of the captured piece square
            # It also checks if the position is in a pseudolegal list of legal moves (because we can have landing squares that aren't in the same diagonal as the staring square (e.g. if the piece is in 28, 31 isn't a possible landing square))
            if down_direction_possible == down_direction and left_direction_possible == left_direction and position in adjacent_positions:
                legal_adjacent_positions.append(self.get_square(row, column))

        enemy_piece.king = was_king
        adjacent_positions = legal_adjacent_positions

        positions_to_check = []
        row_to_check = current_row
        position_to_check = self.position

        for multiplier in range(1, self.board.height):
            row_change = 1 if down_direction else -1
            add = self.squares_per_row
            add *= row_change

            # Because only half the squares are playable

            # Add 1 square if the piece is moving diagonally to the right
            add += int(not left_direction)
            # Remove 1 square if the row the piece is in, is an odd 4
            add -= int(row_to_check % 2 == 1)
            # e.g. From 30 to 25: We start with add=5 because variant="standard".
            # The piece is going to the right (from white's perspective always), so we add 1 to add.
            # The piece is in an odd row (square 30 is in row 5) so we subtract 1 from add.
            # We end with add=5, which is the correct result.

            position_to_check += add
            row_to_check += row_change

            # Checks if position is in the correct row.
            # e.g. If a piece is at square 35 it can't go to the right, but when we subtract 6, we get 29, which would
            # be correct if we could go to the right. To prevent this we check that the expected row matches the actual
            # row. In this example we expect row=5, but we get row=4, so we know that this isn't correct.
            if self.get_row_from_position(position_to_check) == row_to_check:
                positions_to_check.append(position_to_check)
            else:
                break

        # Checks if the piece will encounter any other piece in its path
        for index, position in enumerate(positions_to_check):
            enemy_piece_found = False
            for semi_position in positions_to_check[:index + 1]:
                piece = self.board.searcher.get_piece_by_position(semi_position)
                if piece is not None:
                    # It stops if it meets a piece of the same color or another opponent piece
                    if piece.player == self.player or enemy_piece_found:
                        break
                    else:
                        enemy_piece_found = True
                # Kings in multi-captures can't go over a piece they have captured in that move sequence in most
                # variants. They can in turkish, but pieces in turkish draughts can't move diagonally.
                elif semi_position in captures:
                    break
            else:
                if position in adjacent_positions:
                    positions.append(position)
                continue
            break
        return positions

    def get_orthogonal_multiple_squares_behind_enemy(self, enemy_piece: Piece, captures: List[int]) -> List[int]:
        """Get the orthogonal squares behind the enemy piece if the piece can move more than one square (kings)."""
        positions = []
        current_row = self.get_row()
        current_column = self.get_column()
        enemy_column = enemy_piece.get_column()
        enemy_row = enemy_piece.get_row()

        positions_to_check = []
        row_difference = current_row - enemy_row
        column_difference = current_column - enemy_column

        # Get direction of row change
        add_row = 0
        if row_difference < 0:
            add_row = -1
        elif row_difference > 0:
            add_row = 1

        # Get direction of column change
        add_column = 0
        if column_difference < 0:
            add_column = -1
        elif column_difference > 0:
            add_column = 1

        # Check if the pieces are on the same row or column
        if (column_difference == 0 or row_difference == 0) and (row_difference % 2 == 0 if self.half_of_the_squares_are_playable else True):
            for multiplier in range(1, max(self.board.width, self.board.height)):
                # In frisian, the square directly in front, behind, on the left and on the right
                # of the piece isn't playable.
                if multiplier % 2 == 1 and self.half_of_the_squares_are_playable:
                    continue

                next_row = current_row - multiplier * add_row
                next_column = current_column - multiplier * add_column
                if next_row in self.board.position_layout and next_column in self.board.position_layout[next_row]:
                    positions_to_check.append(self.get_square(next_row, next_column))

                next_row = enemy_row - multiplier * add_row
                next_column = enemy_column - multiplier * add_column
                if next_row in self.board.position_layout and next_column in self.board.position_layout[next_row]:
                    positions.append(self.get_square(next_row, next_column))

            # Kings in multi-captures can't go over a piece they have captured in that move sequence in
            # frisian and frysk!. In turkish they can, but we use the last capture to prevent the piece from
            # turning 180 degrees, which is not allowed.
            if self.kings_can_turn_180_degrees_in_multicapture:
                captures = []
            elif self.kings_can_jump_over_an_already_captured_piece and captures:
                captures = [captures[-1]]

            new_positions = []
            for index, position in enumerate(positions_to_check):
                enemy_piece_found = False
                for semi_position in positions_to_check[:index + 1]:
                    piece = self.board.searcher.get_piece_by_position(semi_position)
                    if piece is not None:
                        # It stops if it meets a piece of the same color or another opponent piece.
                        if piece.player == self.player or enemy_piece_found:
                            break
                        else:
                            enemy_piece_found = True
                    elif semi_position in captures:
                        # Kings in multi-captures can't go over a piece they have captured in that move sequence in
                        # frisian and frysk!. In turkish they can, but we use the last capture to prevent the piece from
                        # turning 180 degrees, which is not allowed.
                        break
                else:
                    if position in positions:
                        new_positions.append(position)
                    continue
                break

            positions = new_positions

        return positions

    def get_position_behind_enemy(self, enemy_piece: Piece, captures: List[int]) -> List[int]:
        """
        Gets the positions the piece can land after capturing the enemy piece.
        :param captures: it contains all the positions of the pieces that this piece has captured during this multi-capture,
        because kings in multi-captures can't go over a piece they have captured in that move sequence in most variants
        """
        positions = []
        if not self.king:
            if self.orthogonal_captures or self.orthogonal_moves:
                positions += self.get_orthogonal_one_square_behind_enemy(enemy_piece)
            if not positions and self.diagonal_moves:
                positions += self.get_diagonal_one_square_behind_enemy(enemy_piece)
        else:
            if self.kings_can_move_more_than_one_square:
                if self.orthogonal_captures or self.orthogonal_moves:
                    positions += self.get_orthogonal_multiple_squares_behind_enemy(enemy_piece, captures)
                if not positions and self.diagonal_moves:
                    positions += self.get_diagonal_multiple_squares_behind_enemy(enemy_piece, captures)
            else:
                if self.orthogonal_captures or self.orthogonal_moves:
                    positions += self.get_orthogonal_one_square_behind_enemy(enemy_piece)
                if not positions and self.diagonal_moves:
                    positions += self.get_diagonal_one_square_behind_enemy(enemy_piece)
        return positions

    def get_possible_positional_moves(self) -> List[List[int]]:
        """Get all possible positional moves (not capture moves) for this piece."""
        if self.possible_positional_moves is None:
            self.possible_positional_moves = self.build_possible_positional_moves()

        return self.possible_positional_moves

    def build_possible_positional_moves(self) -> List[List[int]]:
        """Build all possible positional moves (not capture moves) for this piece."""
        new_positions = list(filter((lambda position: self.board.position_is_open(position)), self.get_adjacent_positions()))

        return self.create_moves_from_new_positions(new_positions)

    def create_moves_from_new_positions(self, new_positions: List[int]) -> List[List[int]]:
        """Create moves ([[cur_pos, end_pos1], [cur_pos, end_pos2]]) given the ending square ([end_pos1, end_pos2])."""
        return [[self.position, new_position] for new_position in new_positions]

    def get_adjacent_positions(self, capture: bool = False) -> List[int]:
        """Get all adjacent positions of the piece."""
        # In some variants men can't capture backwards
        criteria = bool(capture or self.king) if self.men_can_capture_backwards else bool(self.king)
        return self.get_directional_adjacent_positions(forward=True, capture=capture) + (self.get_directional_adjacent_positions(forward=False, capture=capture) if criteria else [])

    def get_column(self) -> int:
        """Get the piece's column."""
        return (self.position - 1) % self.board.width

    def get_row(self) -> int:
        """Get the piece's row."""
        return self.get_row_from_position(self.position)

    def is_on_enemy_home_row(self) -> bool:
        """Get if the piece is on the enemy's home row (used for promotions)."""
        return self.get_row() == self.get_row_from_position(1 if self.other_player == BLACK else self.board.position_count)

    def get_row_from_position(self, position: int) -> int:
        """Get the piece's row, given its square."""
        return ceil(position / self.board.width) - 1

    def get_directional_diagonal_one_square_adjacent_positions(self, forward: bool) -> List[int]:
        """
        Get the diagonal directional adjacent positions if the piece can move only one square
        (usually men, but in some variants kings can also only move one square).
        """
        current_row = self.get_row()
        next_row = current_row + ((1 if self.player == BLACK else -1) * (1 if forward else -1))

        if next_row not in self.board.position_layout:
            return []

        next_column_indexes = self.get_next_column_indexes(current_row, self.get_column())

        return [self.board.position_layout[next_row][column_index] for column_index in next_column_indexes]

    def get_directional_orthogonal_one_square_adjacent_positions(self, forward: bool, capture: bool) -> List[int]:
        """
        Get the orthogonal directional adjacent positions if the piece can move only one square
        (usually men, but in some variants kings can also only move one square).
        """
        positions = []
        current_row = self.get_row()
        current_column = self.get_column()

        # If only half_of_the_squares_are_playable, the first square directly in front is 2 rows away,
        # while if all the squares are playable, it is only 1 row away.
        row_in_front = 2 if self.half_of_the_squares_are_playable else 1

        if self.orthogonal_moves or self.orthogonal_captures and self.men_can_capture_backwards and capture:
            # If self.orthogonal_moves is False:
            # With forward=True we will calculate left and up and with forward=False we will calculate right and down.
            # e.g. If the piece is at 33, square 32 will be considered in forward=True and
            # square 34 will be considered in forward=False.
            # So we will use both forward=True and forward=False, to get all 4 directions.
            # This is used for example in frisian.
            #
            # if self.orthogonal_moves is True:
            # With forward=True we will calculate left, right and up.
            # forward=False will return the same as forward=True because it is not meant to be used when
            # self.orthogonal_moves is True.
            # This is used for example in turkish.

            next_row = current_row + ((row_in_front if self.player == BLACK else -row_in_front) * (1 if forward else -1))
            next_column = current_column + ((1 if self.player == BLACK else -1) * (1 if forward else -1))

            if next_row in self.board.position_layout:
                positions.append(self.board.position_layout[next_row][current_column])
            if next_column in self.board.position_layout[current_row]:
                positions.append(self.board.position_layout[current_row][next_column])

            if self.orthogonal_moves:
                next_column = current_column - ((1 if self.player == BLACK else -1) * (1 if forward else -1))
                if next_column in self.board.position_layout[current_row]:
                    positions.append(self.board.position_layout[current_row][next_column])

        return positions

    def get_directional_diagonal_multiple_squares_adjacent_positions(self, forward: bool, capture: bool) -> List[int]:
        """Get the diagonal directional adjacent positions if the piece can move more than one square (kings)."""
        positions = []
        current_row = self.get_row()
        current_column = self.get_column()

        positions_diagonal_1 = []
        positions_diagonal_2 = []
        for i in range(1, self.board.height):
            next_row = current_row + ((i if self.player == BLACK else -i) * (1 if forward else -1))

            if next_row not in self.board.position_layout:
                continue

            next_column_indexes = self.get_next_column_indexes(current_row, current_column, i, True)
            if 0 <= next_column_indexes[0] < self.board.width:
                positions_diagonal_1.append(self.board.position_layout[next_row][next_column_indexes[0]])
            if 0 <= next_column_indexes[1] < self.board.width:
                positions_diagonal_2.append(self.board.position_layout[next_row][next_column_indexes[1]])

        for index, position in enumerate(positions_diagonal_1):
            for semi_position in positions_diagonal_1[:index + 1]:
                piece = self.board.searcher.get_piece_by_position(semi_position)
                if piece is not None and not capture:
                    # If we encounter a piece, and we are searching for a positional move not a capture move, we break the loop
                    break
            else:
                positions.append(position)
                continue
            break

        for index, position in enumerate(positions_diagonal_2):
            for semi_position in positions_diagonal_2[:index + 1]:
                piece = self.board.searcher.get_piece_by_position(semi_position)
                if piece is not None and not capture:
                    # If we encounter a piece, and we are searching for a positional move not a capture move, we break the loop
                    break
            else:
                positions.append(position)
                continue
            break

        return positions

    def get_directional_orthogonal_multiple_squares_adjacent_positions(self, forward: bool, captures: bool) -> List[int]:
        """Get the orthogonal directional adjacent positions if the piece can move more than one square (kings)."""
        if not (self.orthogonal_moves or self.orthogonal_captures and captures):
            return []
        positions = []
        current_row = self.get_row()
        current_column = self.get_column()
        row_in_front = 2 if self.half_of_the_squares_are_playable else 1

        for multiplier in range(1, self.board.height):
            next_row = current_row + multiplier * (row_in_front if self.player == BLACK else -row_in_front) * (1 if forward else -1)
            if next_row in self.board.position_layout:
                positions.append(self.board.position_layout[next_row][current_column])

        for multiplier in range(1, self.board.width):
            next_column = current_column + multiplier * (1 if self.player == BLACK else -1) * (1 if forward else -1)
            if next_column in self.board.position_layout[current_row]:
                positions.append(self.board.position_layout[current_row][next_column])

        return positions

    def get_directional_adjacent_positions(self, forward: bool, capture: bool = False) -> List[int]:
        """Get the adjacent positions, either forwards or backwards."""
        positions = []
        if not self.king:
            if self.diagonal_moves:
                positions += self.get_directional_diagonal_one_square_adjacent_positions(forward)
            if self.orthogonal_captures:
                positions += self.get_directional_orthogonal_one_square_adjacent_positions(forward, capture)
        else:
            if self.kings_can_move_more_than_one_square:
                if self.diagonal_moves:
                    positions += self.get_directional_diagonal_multiple_squares_adjacent_positions(forward, capture)
                if self.orthogonal_captures:
                    positions += self.get_directional_orthogonal_multiple_squares_adjacent_positions(forward, capture)
            else:
                if self.diagonal_moves:
                    positions += self.get_directional_diagonal_one_square_adjacent_positions(forward)
                if self.orthogonal_captures:
                    positions += self.get_directional_orthogonal_one_square_adjacent_positions(forward, capture)
        return positions

    def get_next_column_indexes(self, current_row: int, current_column: int, i: int = 1, unfiltered: bool = False) -> List[int]:
        """
        Get the index of the next column.
        It isn't as simple as finding the next row (where we only add or subtract 1) but the column index only changes
        once every 2 rows.
        e.g. Square 32 and 27 are both in the second row, but 27 and 31 are not.
        :param i: current_row ± i is the row we should search (the columns for current_row + i are the same as current_row - i).
        e.g. if current row=3 and i=2, we will return the 2 possible columns for row 5 (which will be the same for row 1).
        """
        column_indexes = [0, 0]
        start_right = current_row % 2 == int(self.bottom_left_square_is_playable)
        for semi_i in range(1, i + 1):
            if start_right and semi_i % 2 == 1 or not start_right and semi_i % 2 == 0:
                column_indexes[1] += 1
            else:
                column_indexes[0] -= 1

        column_indexes = list(map(lambda value: value + current_column, column_indexes))

        if unfiltered:
            return column_indexes
        else:
            return list(filter((lambda column_index: 0 <= column_index < self.board.width), column_indexes))

    @property
    def other_player(self) -> int:
        return BLACK if self.player == WHITE else WHITE
